import time

from contextlib import contextmanager

from .sidecar import SidecarSubProcess
from .sidecar_messages import Message, MessageTypes

COUNTER_TYPE = "COUNTER"
GAUGE_TYPE = "GAUGE"
MEASURE_TYPE = "MEASURE"
TIMER_TYPE = "TIMER"

class NullMonitor(object):

    def __init__(self, *args, **kwargs):
        pass

    def start(self):
        pass

    @contextmanager
    def count(self, name):
        yield

    @contextmanager
    def measure(self, name):
        yield

    def gauge(self, gauge):
        pass

    def terminate(self):
        pass

class Monitor(NullMonitor):

    def __init__(self, monitor_type, env, flow_name):
        # type: (str) -> None
        self.sidecar_process = None
        self.monitor_type = monitor_type
        self.env_info = env.get_environment_info()
        self.env_info["flow_name"] = flow_name

    def start(self):
        if self.sidecar_process is None:
            self.sidecar_process = SidecarSubProcess(self.monitor_type)

    @contextmanager
    def count(self, name):
        if self.sidecar_process is not None:
            counter = Counter(name, self.env_info)
            counter.increment()
            payload = {
                'counter': counter.to_dict()
            }
            msg = Message(MessageTypes.LOG_EVENT, payload)
            yield
            self.sidecar_process.msg_handler(msg)
        else:
            yield

    @contextmanager
    def measure(self, name):
        if self.sidecar_process is not None:
            timer = Timer(name + "_timer", self.env_info)
            counter = Counter(name + "_counter", self.env_info)
            timer.start()
            counter.increment()
            yield
            timer.end()
            payload = {
                'counter': counter.to_dict(),
                'timer': timer.to_dict()
            }
            msg = Message(MessageTypes.LOG_EVENT, payload)
            self.sidecar_process.msg_handler(msg)
        else:
            yield

    def gauge(self, gauge):
        if self.sidecar_process is not None:
            payload = {
                'gauge': gauge.to_dict()
            }
            msg = Message(MessageTypes.LOG_EVENT, payload)
            self.sidecar_process.msg_handler(msg)

    def terminate(self):
        if self.sidecar_process is not None:
            self.sidecar_process.kill()


class Metric(object):
    """
        Abstract base class
    """

    def __init__(self, type, env):
        self._env = env
        self._type = type

    @property
    def name(self):
        raise NotImplementedError()

    @property
    def flow_name(self):
        return self._env['flow_name']

    @property
    def env(self):
        return self._env

    @property
    def value(self):
        raise NotImplementedError()

    def set_env(self, env):
        self._env = env

    def to_dict(self):
        return {
            '_env': self._env,
            '_type': self._type,
        }


class Timer(Metric):
    def __init__(self, name, env):
        super(Timer, self).__init__(TIMER_TYPE, env)
        self._name = name
        self._start = 0
        self._end = 0

    @property
    def name(self):
        return self._name

    def start(self):
        self._start = time.time()

    def end(self):
        self._end = time.time()

    def set_start(self, start):
        self._start = start

    def set_end(self, end):
        self._end = end

    def get_duration(self):
        return self._end - self._start

    @property
    def value(self):
        return (self._end - self._start) * 1000

    def to_dict(self):
        parent_dict = super(Timer, self).to_dict()
        parent_dict['_name'] = self.name
        parent_dict['_start'] = self._start
        parent_dict['_end'] = self._end
        return parent_dict


class Counter(Metric):
    def __init__(self, name, env):
        super(Counter, self).__init__(COUNTER_TYPE, env)
        self._name = name
        self._count = 0

    @property
    def name(self):
        return self._name

    def increment(self):
        self._count += 1

    def set_count(self, count):
        self._count = count

    @property
    def value(self):
        return self._count

    def to_dict(self):
        parent_dict = super(Counter, self).to_dict()
        parent_dict['_name'] = self.name
        parent_dict['_count'] = self._count
        return parent_dict


class Gauge(Metric):
    def __init__(self, name, env):
        super(Gauge, self).__init__(GAUGE_TYPE, env)
        self._name = name
        self._value = 0

    @property
    def name(self):
        return self._name

    def set_value(self, val):
        self._value = val

    def increment(self):
        self._value += 1

    @property
    def value(self):
        return self._value

    def to_dict(self):
        parent_dict = super(Gauge, self).to_dict()
        parent_dict['_name'] = self.name
        parent_dict['_value'] = self.value
        return parent_dict


def deserialize_metric(metrics_dict):
    if metrics_dict is None:
        return

    type = metrics_dict.get('_type')
    name = metrics_dict.get('_name')
    if type == COUNTER_TYPE:
        try:
            counter = Counter(name, None)
            counter.set_env(metrics_dict.get('_env'))
        except Exception as ex:
            return

        counter.set_count(metrics_dict.get('_count'))
        return counter
    elif type == TIMER_TYPE:
        timer = Timer(name, None)
        timer.set_start(metrics_dict.get('_start'))
        timer.set_end(metrics_dict.get('_end'))
        timer.set_env(metrics_dict.get('_env'))
        return timer
    elif type == GAUGE_TYPE:
        gauge = Gauge(name, None)
        gauge.set_env(metrics_dict.get('_env'))
        gauge.set_value(metrics_dict.get('_value'))
        return gauge
    else:
        raise NotImplementedError("UNSUPPORTED MESSAGE TYPE IN MONITOR")


def get_monitor_msg_type(msg):
    if msg.payload.get('gauge') is not None:
        return GAUGE_TYPE
    if msg.payload.get('counter') is not None:
        if msg.payload.get('timer') is not None:
            return MEASURE_TYPE
        return COUNTER_TYPE


# -*- coding: utf-8 -*-
# File modified under the Apache Licence 2.0 spec. Modified by Aero Technologies.

from collections import defaultdict, deque
import random
import select
import sys
import time
import hashlib

try:
    unicode
except NameError:
    unicode = str
    basestring = str

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import AWS_SANDBOX_ENABLED

class BatchClient(object):
    def __init__(self):
        from ..aws_client import get_aws_client
        self._client = get_aws_client('batch')

    def active_job_queues(self):
        paginator = self._client.get_paginator('describe_job_queues')
        return (
            queue['jobQueueName']
            for page in paginator.paginate()
            for queue in page['jobQueues']
            if queue['state'] == 'ENABLED' and queue['status'] == 'VALID'
        )

    def unfinished_jobs(self):
        queues = self.active_job_queues()
        return (
            job
            for queue in queues
            for status in ['SUBMITTED', 'PENDING', 'RUNNABLE', 'STARTING', 'RUNNING']
            for page in self._client.get_paginator('list_jobs').paginate(
                jobQueue=queue, jobStatus=status
            )
            for job in page['jobSummaryList']
        )

    def describe_jobs(self, job_ids):
        for jobIds in [job_ids[i:i+100] for i in range(0, len(job_ids), 100)]:
            for jobs in self._client.describe_jobs(jobs=jobIds)['jobs']:
                yield jobs

    def describe_job_queue(self, job_queue):
        paginator = self._client.get_paginator('describe_job_queues').paginate(
            jobQueues=[job_queue], maxResults=1)
        return paginator.paginate()['jobQueues'][0] 

    def job(self):
        return BatchJob(self._client)

    def attach_job(self, job_id):
        job = RunningJob(job_id, self._client)
        return job.update()

    def region(self):
        return self._client._client_config.region_name


class BatchJobException(MetaflowException):
    headline = 'AWS Batch job error'


class BatchJob(object):
    def __init__(self, client):
        self._client = client
        tree = lambda: defaultdict(tree)
        self.payload = tree()

    def execute(self):
        if self._image is None:
            raise BatchJobException(
                'Unable to launch AWS Batch job. No docker image specified.'
            )
        if self._iam_role is None:
            raise BatchJobException(
                'Unable to launch AWS Batch job. No IAM role specified.'
            )
        if 'jobDefinition' not in self.payload:
            self.payload['jobDefinition'] = \
                self._register_job_definition(self._image,
                                              self._iam_role,
                                              self.payload['job_queue'],
                                              self._execution_role,
                                              self._shared_memory,
                                              self._max_swap,
                                              self._swappiness)
        response = self._client.submit_job(**self.payload)
        job = RunningJob(response['jobId'], self._client)
        return job.update()

    def _register_job_definition(self,
                                 image,
                                 job_role,
                                 job_queue,
                                 execution_role,
                                 shared_memory,
                                 max_swap,
                                 swappiness):
        # identify platform from any compute environment associated with the
        # queue
        if AWS_SANDBOX_ENABLED:
            # within the Metaflow sandbox, we can't execute the
            # describe_job_queues directive for AWS Batch to detect compute
            # environment platform, so let's just default to EC2 for now.
            platform = "EC2"
        else:
            response = self._client.describe_job_queues(jobQueues=[job_queue])
            if len(response['jobQueues']) == 0:
                raise BatchJobException(
                    'AWS Batch Job Queue %s not found.' % job_queue)
            compute_environment = response['jobQueues'][0] \
                                    ['computeEnvironmentOrder'][0] \
                                    ['computeEnvironment']
            response = self._client.describe_compute_environments(
                computeEnvironments=[compute_environment])
            platform = response['computeEnvironments'][0] \
                            ['computeResources']['type']

        # compose job definition
        job_definition = {
            'type': 'container',
            'containerProperties': {
                'image': image,
                'jobRoleArn': job_role,
                'command': ['echo', 'hello world'],
                'resourceRequirements': [
                    {
                        'value': '1',
                        'type': 'VCPU'
                    },
                    {
                        'value': '4096',
                        'type': 'MEMORY'
                    }
                ]
            }
        }

        if platform == 'FARGATE' or platform == 'FARGATE_SPOT':
            if execution_role is None:
                raise BatchJobException(
                    'No AWS Fargate task execution IAM role found. Please see '
                    'https://docs.aws.amazon.com/batch/latest/userguide/execution-IAM-role.html '
                    'and set the role as METAFLOW_ECS_FARGATE_EXECUTION_ROLE '
                    'environment variable.')
            job_definition['containerProperties']['executionRoleArn'] = \
                execution_role
            job_definition['platformCapabilities'] = ['FARGATE']
            job_definition['containerProperties']['networkConfiguration'] = \
                {'assignPublicIp': 'ENABLED'}
        
        # if platform == 'EC2' or platform == 'SPOT':
        #     if 'linuxParameters' not in job_definition['containerProperties']:
        #         job_definition['containerProperties']['linuxParameters'] = {}
        #     if shared_memory is not None:
        #         if not (isinstance(shared_memory, (int, unicode, basestring)) and 
        #             int(shared_memory) > 0):
        #             raise BatchJobException(
        #                 'Invalid shared memory size value ({}); '
        #                 'it should be greater than 0'.format(shared_memory))
        #         else:
        #             job_definition['containerProperties'] \
        #                 ['linuxParameters']['sharedMemorySize'] = int(shared_memory)
        #     if swappiness is not None: 
        #         if not (isinstance(swappiness, (int, unicode, basestring)) and 
        #             int(swappiness) >= 0 and int(swappiness) < 100):
        #             raise BatchJobException(
        #                 'Invalid swappiness value ({}); '
        #                 '(should be 0 or greater and less than 100)'.format(swappiness))
        #         else:
        #             job_definition['containerProperties'] \
        #                 ['linuxParameters']['swappiness'] = int(swappiness)
        #     if max_swap is not None: 
        #         if not (isinstance(max_swap, (int, unicode, basestring)) and 
        #             int(max_swap) >= 0):
        #             raise BatchJobException(
        #                 'Invalid swappiness value ({}); '
        #                 '(should be 0 or greater)'.format(max_swap))
        #         else:
        #             job_definition['containerProperties'] \
        #                 ['linuxParameters']['maxSwap'] = int(max_swap)

        # check if job definition already exists
        def_name = 'metaflow_%s' % \
            hashlib.sha224(str(job_definition).encode('utf-8')).hexdigest()
        payload = {'jobDefinitionName': def_name, 'status': 'ACTIVE'}
        response = self._client.describe_job_definitions(**payload)
        if len(response['jobDefinitions']) > 0:
            return response['jobDefinitions'][0]['jobDefinitionArn']

        # else create a job definition
        job_definition['jobDefinitionName'] = def_name
        try:
            response = self._client.register_job_definition(**job_definition)
        except Exception as ex:
            if type(ex).__name__ == 'ParamValidationError' and \
                    (platform == 'FARGATE' or platform == 'FARGATE_SPOT'):
                raise BatchJobException(
                    '%s \nPlease ensure you have installed boto3>=1.16.29 if '
                    'you intend to launch Aero tasks' % ex)
            else:
                raise ex
        return response['jobDefinitionArn']

    def job_def(self,
                image,
                iam_role,
                job_queue,
                execution_role,
                shared_memory,
                max_swap,
                swappiness):
        self.payload['jobDefinition'] = \
            self._register_job_definition(image,
                                          iam_role,
                                          job_queue,
                                          execution_role,
                                          shared_memory,
                                          max_swap,
                                          swappiness)
        return self

    def job_name(self, job_name):
        self.payload['jobName'] = job_name
        return self

    def job_queue(self, job_queue):
        self.payload['jobQueue'] = job_queue
        return self

    def image(self, image):
        self._image = image
        return self

    def iam_role(self, iam_role):
        self._iam_role = iam_role
        return self

    def execution_role(self, execution_role):
        self._execution_role = execution_role
        return self

    def shared_memory(self, shared_memory):
        self._shared_memory = shared_memory
        return self

    def max_swap(self, max_swap):
        self._max_swap = max_swap
        return self

    def swappiness(self, swappiness):
        self._swappiness = swappiness
        return self

    def command(self, command):
        if 'command' not in self.payload['containerOverrides']:
            self.payload['containerOverrides']['command'] = []
        self.payload['containerOverrides']['command'].extend(command)
        return self

    def cpu(self, cpu):
        if not (isinstance(cpu, (int, unicode, basestring, float)) and float(cpu) > 0):
            raise BatchJobException(
                'Invalid CPU value ({}); it should be greater than 0'.format(cpu))
        if 'resourceRequirements' not in self.payload['containerOverrides']:
            self.payload['containerOverrides']['resourceRequirements'] = []
        self.payload['containerOverrides']['resourceRequirements'].append(
            {'value' : str(cpu), 'type': 'VCPU'}
        )
        self.cpu_val = cpu
        return self

    def memory(self, mem):
        if not (isinstance(mem, (int, unicode, basestring)) and int(mem) > 0):
            raise BatchJobException(
                'Invalid memory value ({}); it should be greater than 0'.format(mem))
        if 'resourceRequirements' not in self.payload['containerOverrides']:
            self.payload['containerOverrides']['resourceRequirements'] = []
        self.payload['containerOverrides']['resourceRequirements'].append(
            {'value' : str(mem), 'type': 'MEMORY'}
        )
        self.memory_val = mem
        return self

    def gpu(self, gpu):
        if not (isinstance(gpu, (int, unicode, basestring))):
            raise BatchJobException(
                'invalid gpu value: ({}) (should be 0 or greater)'.format(gpu))
        if int(gpu) > 0:
            if 'resourceRequirements' not in self.payload['containerOverrides']:
                self.payload['containerOverrides']['resourceRequirements'] = []
            self.payload['containerOverrides']['resourceRequirements'].append(
                {'type': 'GPU', 'value': str(gpu)}
            )
        return self

    def environment_variable(self, name, value):
        if 'environment' not in self.payload['containerOverrides']:
            self.payload['containerOverrides']['environment'] = []
        value = str(value)
        if value.startswith("$$.") or value.startswith("$."):
            # Context Object substitution for AWS Step Functions
            # https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html
            self.payload['containerOverrides']['environment'].append(
                {'name': name, 'value.$': value}
            )
        else:
            self.payload['containerOverrides']['environment'].append(
                {'name': name, 'value': value}
            )
        return self

    def timeout_in_secs(self, timeout_in_secs):
        self.payload['timeout']['attemptDurationSeconds'] = timeout_in_secs
        return self

    def parameter(self, key, value):
        self.payload['parameters'][key] = str(value)
        return self

    def attempts(self, attempts):
        self.payload['retryStrategy']['attempts'] = attempts
        return self

class Throttle(object):
    def __init__(self, delta_in_secs=1, num_tries=20):
        self.delta_in_secs = delta_in_secs
        self.num_tries = num_tries
        self._now = None
        self._reset()

    def _reset(self):
        self._tries_left = self.num_tries
        self._wait = self.delta_in_secs

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            now = time.time()
            if self._now is None or (now - self._now > self._wait):
                self._now = now
                try:
                    func(*args, **kwargs)
                    self._reset()
                except TriableException as ex:
                    self._tries_left -= 1
                    if self._tries_left == 0:
                        raise ex.ex
                    self._wait = (self.delta_in_secs*1.2)**(self.num_tries-self._tries_left) + \
                        random.randint(0, 3*self.delta_in_secs)
        return wrapped

class TriableException(Exception):
    def __init__(self, ex):
        self.ex = ex

class RunningJob(object):

    NUM_RETRIES = 8

    def __init__(self, id, client):
        self._id = id
        self._client = client
        self._data = {}

    def __repr__(self):
        return '{}(\'{}\')'.format(self.__class__.__name__, self._id)

    def _apply(self, data):
        self._data = data

    @Throttle()
    def _update(self):
        try:
            data = self._client.describe_jobs(jobs=[self._id])
        except self._client.exceptions.ClientError as err:
            code = err.response['ResponseMetadata']['HTTPStatusCode']
            if code == 429 or code >= 500:
                raise TriableException(err)
            raise err
        # There have been sporadic reports of empty responses to the
        # batch.describe_jobs API call, which can potentially happen if the
        # batch.submit_job API call is not strongly consistent(¯\_(ツ)_/¯).
        # We add a check here to guard against that. The `update()` call
        # will ensure that we poll `batch.describe_jobs` until we get a
        # satisfactory response at least once through out the lifecycle of
        # the job.
        if len(data['jobs']) == 1:
            self._apply(data['jobs'][0])

    def update(self):
        self._update()
        while not self._data:
            self._update()
        return self

    @property
    def id(self):
        return self._id

    @property
    def info(self):
        if not self._data:
            self.update()
        return self._data

    @property
    def job_name(self):
        return self.info['jobName']

    @property
    def job_queue(self):
        return self.info['jobQueue']

    @property
    def status(self):
        if not self.is_done:
            self.update()
        return self.info['status']

    @property
    def status_reason(self):
        return self.info.get('statusReason')

    @property
    def created_at(self):
        return self.info['createdAt']

    @property
    def stopped_at(self):
        return self.info.get('stoppedAt', 0)

    @property
    def is_done(self):
        if self.stopped_at == 0:
            self.update()
        return self.stopped_at > 0

    @property
    def is_running(self):
        return self.status == 'RUNNING'

    @property
    def is_successful(self):
        return self.status == 'SUCCEEDED'

    @property
    def is_crashed(self):
        # TODO: Check statusmessage to find if the job crashed instead of failing
        return self.status == 'FAILED'

    @property
    def reason(self):
        return self.info['container'].get('reason')

    @property
    def status_code(self):
        if not self.is_done:
            self.update()
        return self.info['container'].get('exitCode')

    def wait_for_running(self):
        if not self.is_running and not self.is_done:
            BatchWaiter(self._client).wait_for_running(self.id)

    @property
    def log_stream_name(self):
        return self.info['container'].get('logStreamName')

    def logs(self):
        def get_log_stream(job):
            log_stream_name = job.log_stream_name
            if log_stream_name:
                return BatchLogs('/aws/batch/job', log_stream_name, sleep_on_no_data=1)
            else:
                return None

        log_stream = None
        while True:
            if self.is_running or self.is_done or self.is_crashed:
                log_stream = get_log_stream(self)
                break
            elif not self.is_done:
                self.wait_for_running()

        if log_stream is None:
            return 
        exception = None
        for i in range(self.NUM_RETRIES + 1):
            try:
                check_after_done = 0
                for line in log_stream:
                    if not line:
                        if self.is_done:
                            if check_after_done > 1:
                                return
                            check_after_done += 1
                        else:
                            pass
                    else:
                        i = 0
                        yield line
                return
            except Exception as ex:
                exception = ex
                if self.is_crashed:
                    break
                #sys.stderr.write(repr(ex) + '\n')
                if i < self.NUM_RETRIES:
                    time.sleep(2 ** i + random.randint(0, 5))
        raise BatchJobException(repr(exception))

    def kill(self):
        if not self.is_done:
            self._client.terminate_job(
                jobId=self._id, reason='Metaflow initiated job termination.')
        return self.update()


class BatchWaiter(object):
    def __init__(self, client):
        try:
            from botocore import waiter
        except:
            raise BatchJobException(
                'Could not import module \'botocore\' which '
                'is required for Batch jobs. Install botocore '
                'first.'
            )
        self._client = client
        self._waiter = waiter

    def wait_for_running(self, job_id):
        model = self._waiter.WaiterModel(
            {
                'version': 2,
                'waiters': {
                    'JobRunning': {
                        'delay': 1,
                        'operation': 'DescribeJobs',
                        'description': 'Wait until job starts running',
                        'maxAttempts': 1000000,
                        'acceptors': [
                            {
                                'argument': 'jobs[].status',
                                'expected': 'SUCCEEDED',
                                'matcher': 'pathAll',
                                'state': 'success',
                            },
                            {
                                'argument': 'jobs[].status',
                                'expected': 'FAILED',
                                'matcher': 'pathAny',
                                'state': 'success',
                            },
                            {
                                'argument': 'jobs[].status',
                                'expected': 'RUNNING',
                                'matcher': 'pathAny',
                                'state': 'success',
                            },
                        ],
                    }
                },
            }
        )
        self._waiter.create_waiter_with_client('JobRunning', model, self._client).wait(
            jobs=[job_id]
        )

class BatchLogs(object):
    def __init__(self, group, stream, pos=0, sleep_on_no_data=0):
        from ..aws_client import get_aws_client
        self._client = get_aws_client('logs')
        self._group = group
        self._stream = stream
        self._pos = pos
        self._sleep_on_no_data = sleep_on_no_data
        self._buf = deque()
        self._token = None

    def _get_events(self):
        try:
            if self._token:
                response = self._client.get_log_events(
                    logGroupName=self._group,
                    logStreamName=self._stream,
                    startTime=self._pos,
                    nextToken=self._token,
                    startFromHead=True,
                )
            else:
                response = self._client.get_log_events(
                    logGroupName=self._group,
                    logStreamName=self._stream,
                    startTime=self._pos,
                    startFromHead=True,
                )
            self._token = response['nextForwardToken']
            return response['events']
        except self._client.exceptions.ResourceNotFoundException as e:
            # The logs might be delayed by a bit, so we can simply try
            # again next time.
            return []

    def __iter__(self):
        while True:
            self._fill_buf()
            if len(self._buf) == 0:
                yield ''
                if self._sleep_on_no_data > 0:
                    select.poll().poll(self._sleep_on_no_data * 1000)
            else:
                while self._buf:
                    yield self._buf.popleft()

    def _fill_buf(self):
        events = self._get_events()
        for event in events:
            self._buf.append(event['message'])
            self._pos = event['timestamp']


import os
import json
import random
import string
import zlib
from itertools import dropwhile, islice

from metaflow.util import to_bytes

def _token_generator(token_prefix):
    for i in range(10000):
        prefix = '%s-%d-' % (token_prefix, i)
        # we need to use a consistent hash here, which is why
        # random.seed(prefix) or random.seed(hash(prefix)) won't work
        random.seed(zlib.adler32(to_bytes(prefix)))
        yield prefix + ''.join(random.sample(string.ascii_lowercase, 4))

def _makedirs(path):
    # this is for python2 compatibility.
    # Python3 has os.makedirs(exist_ok=True).
    try:
        os.makedirs(path)
    except OSError as x:
        if x.errno == 17:
            return
        else:
            raise

def _load_config(path):
    if os.path.exists(path):
        with open(path) as f:
            return json.load(f)
    else:
        return {}

def _path(token_prefix):
    home = os.environ.get('METAFLOW_HOME', '~/.metaflowconfig')
    return os.path.expanduser('%s/%s' % (home, token_prefix))

def new_token(token_prefix, prev_token=None):
    if prev_token is None:
        for token in _token_generator(token_prefix):
            return token
    else:
        it = dropwhile(lambda x: x != prev_token,
                       _token_generator(token_prefix))
        for _ in it:
            return next(it)
        else:
            return None

def load_token(token_prefix):
    config = _load_config(_path(token_prefix))
    return config.get('production_token')

def store_token(token_prefix, token):
    path = _path(token_prefix)
    config = _load_config(path)
    config['production_token'] = token
    _makedirs(os.path.dirname(path))
    with open(path, 'w') as f:
        json.dump(config, f)


# File modified under the Apache Licence 2.0 spec. Modified by Aero Technologies.
# Altered how profile is set based on local/remote execution
import logging
import os
from pathlib import Path
from metaflow.metaflow_config import AERO_ID_TOKEN, AERO_IDENTITY_POOL, AERO_PROVIDER


def create_client_credentials():
    import boto3

    # Catch if running on Batch
    if 'METAFLOW_INPUT_PATHS_0' in os.environ or 'MANAGED_BY_AWS' in os.environ:
        return boto3.session.Session(
            region_name='eu-west-1'
        )

    client = boto3.client('cognito-identity', region_name='eu-west-1')

    try:
        identity_response = client.get_id(
                                IdentityPoolId=AERO_IDENTITY_POOL, 
                                Logins = {AERO_PROVIDER: AERO_ID_TOKEN})
    except Exception as e:
        raise Exception("Credentials have expired, please run 'aero account login' again")

    identity_id = identity_response['IdentityId']

    resp = client.get_credentials_for_identity(
                    IdentityId=identity_id,
                    Logins={AERO_PROVIDER: AERO_ID_TOKEN})

    secret_key = resp['Credentials']['SecretKey']
    access_key = resp['Credentials']['AccessKeyId']
    session_token = resp['Credentials']['SessionToken']

    session = boto3.session.Session(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token,
        region_name='eu-west-1'
    )

    try:
        sts = session.client('sts')
        sts.get_caller_identity()
    except:
        raise Exception("Credentials have expired, please run 'aero account login' again")

    return session

AWS_SESSION = create_client_credentials()

def get_aws_client(module, with_error=False, params={}):
    from metaflow.exception import MetaflowException  

    try:
        import boto3
        from botocore.exceptions import ClientError
    except (NameError, ImportError):
        raise MetaflowException(
            "Could not import module 'boto3'. Install boto3 first.")

    if with_error:
        return AWS_SESSION.client(module, **params), ClientError
    return AWS_SESSION.client(module, **params)



import os
import sys
import time
import subprocess
from threading import Thread

from metaflow.metaflow_profile import profile
from metaflow.sidecar import SidecarSubProcess
from . import update_delay, BASH_SAVE_LOGS_ARGS

class SaveLogsPeriodically(object):

    def __init__(self):
        self._sidecar = SidecarSubProcess("save_logs_periodically")

    def shutdown(self):
        self._sidecar.kill()

class SaveLogsPeriodicallySidecar(object):

    def __init__(self):
        self._thread = Thread(target=self._update_loop)
        self.is_alive = True
        self._thread.start()

    def process_message(self, msg):
        pass

    def shutdown(self):
        self.is_alive = False

    def _update_loop(self):
        def _file_size(path):
            if os.path.exists(path):
                return os.path.getsize(path)
            else:
                return 0

        # these env vars are set by mflog.mflog_env
        FILES = [os.environ['MFLOG_STDOUT'], os.environ['MFLOG_STDERR']]
        start_time = time.time()
        sizes = [0 for _ in FILES]
        while self.is_alive:
            new_sizes = list(map(_file_size, FILES))
            if new_sizes != sizes:
                sizes = new_sizes
                try:
                    subprocess.call(BASH_SAVE_LOGS_ARGS)
                except:
                    pass
            time.sleep(update_delay(time.time() - start_time))




from __future__ import print_function
import sys
import os
import time

from .metaflow_config import MAX_ATTEMPTS
from .metadata import MetaDatum
from .datastore import Inputs, MetaflowDatastoreSet
from .exception import MetaflowInternalError,\
    MetaflowDataMissing,\
    MetaflowExceptionWrapper
from .util import all_equal,\
    get_username,\
    resolve_identity
from .current import current

from collections import namedtuple

ForeachFrame = namedtuple('ForeachFrame',
                          ['step', 'var', 'num_splits', 'index'])


class MetaflowTask(object):
    """
    MetaflowTask prepares a Flow instance for execution of a single step.
    """

    def __init__(self,
                 flow,
                 datastore,
                 metadata,
                 environment,
                 console_logger,
                 event_logger,
                 monitor):
        self.flow = flow
        self.datastore = datastore
        self.metadata = metadata
        self.environment = environment
        self.console_logger = console_logger
        self.event_logger = event_logger
        self.monitor = monitor

    def _exec_step_function(self, step_function, input_obj=None):
        self.environment.validate_environment(echo=self.console_logger)
        if input_obj is None:
            step_function()
        else:
            step_function(input_obj)

    def _init_parameters(self, parameter_ds):
        # overwrite Parameters in the flow object
        vars = []
        for var, param in self.flow._get_parameters():
            # make the parameter a read-only property
            # note x=x binds the current value of x to the closure
            def property_setter(
                    _, cls=self.flow.__class__, param=param, var=var, parameter_ds=parameter_ds):
                v = param.load_parameter(parameter_ds[var])
                setattr(cls, var, property(fget=lambda _, val=v: val))
                return v

            setattr(self.flow.__class__, var,
                    property(fget=property_setter))
            vars.append(var)
        self.flow._datastore.passdown_partial(parameter_ds, vars)
        return vars

    def _init_data(self, run_id, join_type, input_paths):
        # We prefer to use the parallelized version to initialize datastores
        # (via MetaflowDatastoreSet) only with more than 4 datastores, because
        # the baseline overhead of using the set is ~1.5s and each datastore
        # init takes ~200-300ms when run sequentially.
        if len(input_paths) > 4:
            prefetch_data_artifacts = None
            if join_type and join_type == 'foreach':
                # Prefetch 'foreach' related artifacts to improve time taken by
                # _init_foreach.
                prefetch_data_artifacts = \
                    ['_foreach_stack', '_foreach_num_splits', '_foreach_var']
            # Note: Specify `pathspecs` while creating the datastore set to
            # guarantee strong consistency and guard against missing input.
            datastore_set = \
                MetaflowDatastoreSet(self.datastore,
                                     self.flow.name,
                                     run_id,
                                     pathspecs=input_paths,
                                     metadata=self.metadata,
                                     event_logger=self.event_logger,
                                     monitor=self.monitor,
                                     prefetch_data_artifacts=prefetch_data_artifacts)
            ds_list = [ds for ds in datastore_set]
            if len(ds_list) != len(input_paths):
                raise MetaflowDataMissing("Some input datastores are missing. "
                                          "Expected: %d Actual: %d" %
                                          (len(input_paths), len(ds_list)))
        else:
            # initialize directly in the single input case.
            ds_list = []
            for input_path in input_paths:
                run_id, step_name, task_id = input_path.split('/')
                ds_list.append(
                    self.datastore(self.flow.name,
                                   run_id=run_id,
                                   step_name=step_name,
                                   task_id=task_id,
                                   metadata=self.metadata,
                                   event_logger=self.event_logger,
                                   monitor=self.monitor))
        if not ds_list:
            # this guards against errors in input paths
            raise MetaflowDataMissing("Input paths *%s* resolved to zero "
                                      "inputs" % ','.join(input_paths))
        return ds_list

    def _init_foreach(self, step_name, join_type, inputs, split_index):
        # these variables are only set by the split step in the output
        # data. They don't need to be accessible in the flow.
        self.flow._foreach_var = None
        self.flow._foreach_num_splits = None

        # There are three cases that can alter the foreach state:
        # 1) start - initialize an empty foreach stack
        # 2) join - pop the topmost frame from the stack
        # 3) step following a split - push a new frame in the stack

        # case 1) - reset the stack
        if step_name == 'start':
            self.flow._foreach_stack = []

        # case 2) - this is a join step
        elif join_type:
            # assert the lineage of incoming branches
            def lineage():
                for i in inputs:
                    if join_type == 'foreach':
                        top = i['_foreach_stack'][-1]
                        bottom = i['_foreach_stack'][:-1]
                        # the topmost indices in the stack are all
                        # different naturally, so ignore them in the
                        # assertion
                        yield bottom + [top._replace(index=0)]
                    else:
                        yield i['_foreach_stack']

            if not all_equal(lineage()):
                raise MetaflowInternalError("Step *%s* tried to join branches "
                                            "whose lineages don't match."
                                            % step_name)

            # assert that none of the inputs are splits - we don't
            # allow empty foreaches (joins immediately following splits)
            if any(not i.is_none('_foreach_var') for i in inputs):
                raise MetaflowInternalError("Step *%s* tries to join a foreach "
                                            "split with no intermediate steps."
                                            % step_name)

            inp = inputs[0]
            if join_type == 'foreach':
                # Make sure that the join got all splits as its inputs.
                # Datastore.resolve() leaves out all undone tasks, so if
                # something strange happened upstream, the inputs list
                # may not contain all inputs which should raise an exception
                stack = inp['_foreach_stack']
                if len(inputs) != stack[-1].num_splits:
                    raise MetaflowDataMissing("Foreach join *%s* expected %d "
                                              "splits but only %d inputs were "
                                              "found" % (step_name,
                                                         stack[-1].num_splits,
                                                         len(inputs)))
                # foreach-join pops the topmost frame from the stack
                self.flow._foreach_stack = stack[:-1]
            else:
                # a non-foreach join doesn't change the stack
                self.flow._foreach_stack = inp['_foreach_stack']

        # case 3) - our parent was a split. Initialize a new foreach frame.
        elif not inputs[0].is_none('_foreach_var'):
            if len(inputs) != 1:
                raise MetaflowInternalError("Step *%s* got multiple inputs "
                                            "although it follows a split step."
                                            % step_name)

            if split_index is None:
                raise MetaflowInternalError("Step *%s* follows a split step "
                                            "but no split_index is "
                                            "specified." % step_name)

            # push a new index after a split to the stack
            frame = ForeachFrame(step_name,
                                 inputs[0]['_foreach_var'],
                                 inputs[0]['_foreach_num_splits'],
                                 split_index)

            stack = inputs[0]['_foreach_stack']
            stack.append(frame)
            self.flow._foreach_stack = stack

    def _clone_flow(self, datastore):
        x = self.flow.__class__(use_cli=False)
        x._set_datastore(datastore)
        return x

    def clone_only(self, step_name, run_id, task_id, clone_origin_task):
        if not clone_origin_task:
            raise MetaflowInternalError("task.clone_only needs a valid "
                                        "clone_origin_task value.")
        # 1. initialize output datastore
        output = self.datastore(self.flow.name,
                                run_id=run_id,
                                step_name=step_name,
                                task_id=task_id,
                                mode='w',
                                metadata=self.metadata,
                                attempt=0,
                                event_logger=self.event_logger,
                                monitor=self.monitor)
        output.init_task()
        origin_run_id, origin_step_name, origin_task_id =\
            clone_origin_task.split('/')
        # 2. initialize origin datastore
        origin = self.datastore(self.flow.name,
                                run_id=origin_run_id,
                                step_name=origin_step_name,
                                task_id=origin_task_id,
                                metadata=self.metadata,
                                event_logger=self.event_logger,
                                monitor=self.monitor)
        output.clone(origin)
        output.done()

    def run_step(self,
                 step_name,
                 run_id,
                 task_id,
                 origin_run_id,
                 input_paths,
                 split_index,
                 retry_count,
                 max_user_code_retries):

        if run_id and task_id:
            self.metadata.register_run_id(run_id)
            self.metadata.register_task_id(run_id, step_name, task_id)
        else:
            raise MetaflowInternalError("task.run_step needs a valid run_id "
                                        "and task_id")

        if retry_count >= MAX_ATTEMPTS:
            # any results with an attempt ID >= MAX_ATTEMPTS will be ignored
            # by datastore, so running a task with such a retry_could would
            # be pointless and dangerous
            raise MetaflowInternalError("Too many task attempts (%d)! "
                                        "MAX_ATTEMPTS exceeded." % retry_count)

        self.metadata.register_metadata(run_id,
                                        step_name,
                                        task_id,
                                        [MetaDatum(field='attempt',
                                                   value=str(retry_count),
                                                   type='attempt',
                                                   tags=[]),
                                         MetaDatum(field='origin-run-id',
                                                   value=str(origin_run_id),
                                                   type='origin-run-id',
                                                   tags=[]),
                                         MetaDatum(field='ds-type',
                                                   value=self.datastore.TYPE,
                                                   type='ds-type',
                                                   tags=[]),
                                         MetaDatum(field='ds-root',
                                                   value=self.datastore.datastore_root,
                                                   type='ds-root',
                                                   tags=[])])

        step_func = getattr(self.flow, step_name)
        node = self.flow._graph[step_name]
        join_type = None
        if node.type == 'join':
            join_type = self.flow._graph[node.split_parents[-1]].type

        # 1. initialize output datastore
        output = self.datastore(self.flow.name,
                                run_id=run_id,
                                step_name=step_name,
                                task_id=task_id,
                                mode='w',
                                metadata=self.metadata,
                                attempt=retry_count,
                                event_logger=self.event_logger,
                                monitor=self.monitor)
        output.init_task()

        if input_paths:
            # 2. initialize input datastores
            inputs = self._init_data(run_id, join_type, input_paths)

            # 3. initialize foreach state
            self._init_foreach(step_name, join_type, inputs, split_index)

        # 4. initialize the current singleton
        current._set_env(flow_name=self.flow.name,
                         run_id=run_id,
                         step_name=step_name,
                         task_id=task_id,
                         retry_count=retry_count,
                         origin_run_id=origin_run_id,
                         namespace=resolve_identity(),
                         username=get_username(),
                         is_running=True)

        # 5. run task
        output.save_metadata('task_begin', {
            'code_package_sha': os.environ.get('METAFLOW_CODE_SHA'),
            'code_package_ds': os.environ.get('METAFLOW_CODE_DS'),
            'code_package_url': os.environ.get('METAFLOW_CODE_URL'),
            'retry_count': retry_count
        })
        logger = self.event_logger
        start = time.time()
        self.metadata.start_task_heartbeat(self.flow.name, run_id, step_name,
                                           task_id)
        try:
            # init side cars
            logger.start()

            msg = {
                "task_id": task_id,
                "msg": 'task starting',
                "step_name": step_name,
                "run_id": run_id,
                "flow_name": self.flow.name,
                "ts": round(time.time())
            }
            logger.log(msg)

            self.flow._current_step = step_name
            self.flow._success = False
            self.flow._task_ok = None
            self.flow._exception = None
            # Note: All internal flow attributes (ie: non-user artifacts)
            # should either be set prior to running the user code or listed in
            # FlowSpec._EPHEMERAL to allow for proper merging/importing of
            # user artifacts in the user's step code.
            decorators = step_func.decorators
            for deco in decorators:

                deco.task_pre_step(step_name,
                                   output,
                                   self.metadata,
                                   run_id,
                                   task_id,
                                   self.flow,
                                   self.flow._graph,
                                   retry_count,
                                   max_user_code_retries)

                # decorators can actually decorate the step function,
                # or they can replace it altogether. This functionality
                # is used e.g. by catch_decorator which switches to a
                # fallback code if the user code has failed too many
                # times.
                step_func = deco.task_decorate(step_func,
                                               self.flow,
                                               self.flow._graph,
                                               retry_count,
                                               max_user_code_retries)

            if join_type:
                # Join step:

                # Ensure that we have the right number of inputs. The
                # foreach case is checked above.
                if join_type != 'foreach' and\
                   len(inputs) != len(node.in_funcs):
                    raise MetaflowDataMissing("Join *%s* expected %d "
                                              "inputs but only %d inputs "
                                              "were found"
                                              % (step_name,
                                                 len(node.in_funcs),
                                                 len(inputs)))

                # Multiple input contexts are passed in as an argument
                # to the step function.
                input_obj = Inputs(self._clone_flow(inp) for inp in inputs)
                self.flow._set_datastore(output)
                # initialize parameters (if they exist)
                # We take Parameter values from the first input,
                # which is always safe since parameters are read-only
                current._update_env({'parameter_names': self._init_parameters(inputs[0])})
                self._exec_step_function(step_func, input_obj)
            else:
                # Linear step:
                # We are running with a single input context.
                # The context is embedded in the flow.
                if len(inputs) > 1:
                    # This should be captured by static checking but
                    # let's assert this again
                    raise MetaflowInternalError("Step *%s* is not a join "
                                                "step but it gets multiple "
                                                "inputs." % step_name)
                self.flow._set_datastore(inputs[0])
                if input_paths:
                    # initialize parameters (if they exist)
                    # We take Parameter values from the first input,
                    # which is always safe since parameters are read-only
                    current._update_env({'parameter_names': self._init_parameters(inputs[0])})
                self._exec_step_function(step_func)

            for deco in decorators:
                deco.task_post_step(step_name,
                                    self.flow,
                                    self.flow._graph,
                                    retry_count,
                                    max_user_code_retries)

            self.flow._task_ok = True
            self.flow._success = True

        except Exception as ex:

            tsk_msg = {
                "task_id": task_id,
                "exception_msg": str(ex),
                "msg": 'task failed with exception',
                "step_name": step_name,
                "run_id": run_id,
                "flow_name": self.flow.name
            }
            logger.log(tsk_msg)

            exception_handled = False
            for deco in decorators:
                res = deco.task_exception(ex,
                                          step_name,
                                          self.flow,
                                          self.flow._graph,
                                          retry_count,
                                          max_user_code_retries)
                exception_handled = bool(res) or exception_handled

            if exception_handled:
                self.flow._task_ok = True
            else:
                self.flow._task_ok = False
                self.flow._exception = MetaflowExceptionWrapper(ex)
                print('%s failed:' % self.flow, file=sys.stderr)
                raise

        finally:
            end = time.time() - start

            msg = {
                "task_id": task_id,
                "msg": 'task ending',
                "step_name": step_name,
                "run_id": run_id,
                "flow_name": self.flow.name,
                "ts": round(time.time()),
                "runtime": round(end)
            }
            logger.log(msg)

            attempt_ok = str(bool(self.flow._task_ok))
            self.metadata.register_metadata(run_id,
                                            step_name,
                                            task_id,
                                            [MetaDatum(field='attempt_ok',
                                                       value=attempt_ok,
                                                       type='internal_attempt_status',
                                                       tags=["attempt_id:{0}".
                                                       format(str(retry_count))
                                                             ])
                                             ])

            output.save_metadata('task_end', {})
            output.persist(self.flow)

            # terminate side cars
            logger.terminate()
            self.metadata.stop_heartbeat()

            # this writes a success marker indicating that the
            # "transaction" is done
            output.done()

            # final decorator hook: The task results are now
            # queryable through the client API / datastore
            for deco in decorators:
                deco.task_finished(step_name,
                                   self.flow,
                                   self.flow._graph,
                                   self.flow._task_ok,
                                   retry_count,
                                   max_user_code_retries)



from metaflow.decorators import FlowDecorator


class ScheduleDecorator(FlowDecorator):
    name = 'schedule'
    defaults = {'cron': None,
                'weekly': False,
                'daily': True,
                'hourly': False}

    def flow_init(self, flow, graph,  environment, datastore, logger, echo, options):
        # Currently supports quartz cron expressions in UTC as defined in
        # https://docs.aws.amazon.com/eventbridge/latest/userguide/scheduled-events.html#cron-expressions
        if self.attributes['cron']:
            self.schedule = self.attributes['cron']
        elif self.attributes['weekly']:
            self.schedule = '0 0 ? * SUN *'
        elif self.attributes['hourly']:
            self.schedule = '0 * * * ? *'
        elif self.attributes['daily']:
            self.schedule = '0 0 * * ? *'
        else:
            self.schedule = None


import json
from enum import IntEnum

# Define message enums
class MessageTypes(IntEnum):
    SHUTDOWN = 1
    LOG_EVENT = 2


class Message(object):

    def __init__(self, msg_type, payload):
        self.msg_type = msg_type
        self.payload = payload

    def serialize(self):
        msg = {
            'msg_type': self.msg_type,
            'payload': self.payload,
        }
        return json.dumps(msg)+"\n"


def deserialize(json_msg):
    parsed_json_msg = json.loads(json_msg)
    return Message(MessageTypes(parsed_json_msg['msg_type']),
                   parsed_json_msg['payload'])




from __future__ import print_function

import subprocess
import fcntl
import select
import os
import sys
import platform

from fcntl import F_SETFL
from os import O_NONBLOCK

from .sidecar_messages import Message, MessageTypes
from .debug import debug

MESSAGE_WRITE_TIMEOUT_IN_MS = 1000

NULL_SIDECAR_PREFIX = "nullSidecar"

# for python 2 compatibility
try:
    blockingError = BlockingIOError
except:
    blockingError = OSError

class PipeUnavailableError(Exception):
    """raised when unable to write to pipe given allotted time"""
    pass


class NullSidecarError(Exception):
    """raised when tyring to poll or interact with the fake subprocess in the null sidecar"""
    pass


class MsgTimeoutError(Exception):
    """raised when tyring unable to send message to sidecar in allocated time"""
    pass


class NullPoller(object):
    def poll(self, timeout):
        raise NullSidecarError()


class SidecarSubProcess(object):

    def __init__(self, worker_type):
        # type: (str) -> None
        self.__worker_type = worker_type
        self.__process = None
        self.__poller = None
        self.start()

    def start(self):

        if (self.__worker_type is not None and \
                self.__worker_type.startswith(NULL_SIDECAR_PREFIX)) or \
                (platform.system() == 'Darwin' and sys.version_info < (3, 0)):
                # if on darwin and running python 2 disable sidecars
                # there is a bug with importing poll from select in some cases
                self.__poller = NullPoller()

        else:
            from select import poll
            python_version = sys.executable
            cmdline = [python_version,
                       '-u',
                       os.path.dirname(__file__) + '/sidecar_worker.py',
                       self.__worker_type]
            debug.sidecar_exec(cmdline)

            self.__process = self.__start_subprocess(cmdline)

            if self.__process is not None:
                fcntl.fcntl(self.__process.stdin, F_SETFL, O_NONBLOCK)
                self.__poller = poll()
                self.__poller.register(self.__process.stdin.fileno(),
                                       select.POLLOUT)
            else:
                # unable to start subprocess, fallback to Null sidecar
                self.logger("unable to start subprocess for sidecar %s"
                      % self.__worker_type)
                self.__poller = NullPoller()

    def __start_subprocess(self, cmdline):
        for i in range(3):
            try:
                return subprocess.Popen(cmdline,
                                        stdin=subprocess.PIPE,
                                        stdout=open(os.devnull, 'w'),
                                        bufsize=0)
            except blockingError as be:
                self.logger("warning: sidecar popen failed: %s" % repr(be))
            except Exception as e:
                self.logger(repr(e))
                break

    def kill(self):
        try:
            msg = Message(MessageTypes.SHUTDOWN, None)
            self.emit_msg(msg)
        except:
            pass

    def emit_msg(self, msg):
        msg_ser = msg.serialize().encode('utf-8')
        written_bytes = 0
        while written_bytes < len(msg_ser):
            try:
                fds = self.__poller.poll(MESSAGE_WRITE_TIMEOUT_IN_MS)
                if fds is None or len(fds) == 0:
                    raise MsgTimeoutError("poller timed out")
                for fd, event in fds:
                    if event & select.POLLERR:
                        raise PipeUnavailableError('pipe unavailable')
                    f = os.write(fd, msg_ser[written_bytes:])
                    written_bytes += f
            except NullSidecarError:
                # sidecar is disabled, ignore all messages
                break

    def msg_handler(self, msg, retries=3):
        try:
            self.emit_msg(msg)
        except MsgTimeoutError:
            # drop message, do not retry on timeout
            self.logger("unable to send message due to timeout")
        except Exception as ex:
            if isinstance(ex, PipeUnavailableError):
                self.logger("restarting sidecar %s" % self.__worker_type)
                self.start()
            if retries > 0:
                self.logger("retrying msg send to sidecar")
                self.msg_handler(msg, retries-1)
            else:
                self.logger("error sending log message")
                self.logger(repr(ex))

    def logger(self, msg):
        print("metaflow logger: " + msg, file=sys.stderr)



import base64
import json
from hashlib import sha1

from metaflow.util import to_bytes, to_unicode

class EventBridgeClient(object):

    def __init__(self, name):
        from ..aws_client import get_aws_client
        self._client = get_aws_client('events')
        self.name = format(name)

    def cron(self, cron):
        self.cron = cron
        return self

    def role_arn(self, role_arn):
        self.role_arn = role_arn
        return self

    def state_machine_arn(self, state_machine_arn):
        self.state_machine_arn = state_machine_arn
        return self

    def schedule(self):
        if not self.cron:
            # reset the schedule
            self._disable()
        else:
            self._set()
        return self.name

    def _disable(self):
        try:
            self._client.disable_rule(
                Name=self.name
            )
        except self._client.exceptions.ResourceNotFoundException:
            pass

    def _set(self):
        # Generate a new rule or update existing rule.
        self._client.put_rule(
            Name=self.name,
            ScheduleExpression='cron(%s)' % self.cron,
            Description='Metaflow generated rule for %s' % self.name,
            State='ENABLED'
        )
        # Assign AWS Step Functions ARN to the rule as a target.
        self._client.put_targets(
            Rule=self.name,
            Targets=[
                {
                    'Id':self.name,
                    'Arn':self.state_machine_arn,
                    # Set input parameters to empty.
                    'Input':json.dumps({'Parameters':json.dumps({})}),
                    'RoleArn':self.role_arn
                }
            ]
        )

def format(name):
    # AWS Event Bridge has a limit of 64 chars for rule names.
    # We truncate the rule name if the computed name is greater
    # than 64 chars and append a hashed suffix to ensure uniqueness.
    if len(name) > 64:
        name_hash = to_unicode(
                        base64.b32encode(
                            sha1(to_bytes(name)).digest()))[:16].lower()
        # construct an 64 character long rule name
        return '%s-%s' % (name[:47], name_hash)
    else:
        return name


import sys
# File modified under the Apache Licence 2.0 spec. Modified by Aero Technologies.

try:
    import metaflow_custom.plugins as ext_plugins
except ImportError as e:
    ver = sys.version_info[0] * 10 + sys.version_info[1]
    if ver >= 36:
        # e.path is not None if the error stems from some other place than here
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        if not (isinstance(e, ModuleNotFoundError) and e.path is None):
            print(
                "Cannot load metaflow_custom plugins -- "
                "if you want to ignore, uninstall metaflow_custom package")
            raise

    class _fake(object):
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

        def get_plugin_cli(self):
            return []

    ext_plugins = _fake(
        FLOW_DECORATORS=[],
        STEP_DECORATORS=[],
        ENVIRONMENTS=[],
        METADATA_PROVIDERS=[],
        SIDECARS={},
        LOGGING_SIDECARS={},
        MONITOR_SIDECARS={})


def get_plugin_cli():
    # it is important that CLIs are not imported when
    # __init__ is imported. CLIs may use e.g.
    # parameters.add_custom_parameters which requires
    # that the flow is imported first

    # Add new CLI commands in this list
    from . import package_cli
    from .aws.batch import batch_cli
    from .aws.step_functions import step_functions_cli

    return ext_plugins.get_plugin_cli() + [
        package_cli.cli,
        batch_cli.cli,
        step_functions_cli.cli]


def _merge_lists(base, overrides, attr):
    # Merge two lists of classes by comparing them for equality using 'attr'.
    # This function prefers anything in 'overrides'. In other words, if a class
    # is present in overrides and matches (according to the equality criterion) a class in
    # base, it will be used instead of the one in base.
    l = list(overrides)
    existing = set([getattr(o, attr) for o in overrides])
    l.extend([d for d in base if getattr(d, attr) not in existing])
    return l

# Add new decorators in this list
from .catch_decorator import CatchDecorator
from .timeout_decorator import TimeoutDecorator
from .environment_decorator import EnvironmentDecorator
from .retry_decorator import RetryDecorator
from .aws.batch.batch_decorator import BatchDecorator, ResourcesDecorator
from .aws.step_functions.step_functions_decorator import StepFunctionsInternalDecorator
from .conda.conda_step_decorator import CondaStepDecorator

STEP_DECORATORS = _merge_lists([CatchDecorator,
                                TimeoutDecorator,
                                EnvironmentDecorator,
                                ResourcesDecorator,
                                RetryDecorator,
                                BatchDecorator,
                                StepFunctionsInternalDecorator,
                                CondaStepDecorator], ext_plugins.STEP_DECORATORS, 'name')

# Add Conda environment
from .conda.conda_environment import CondaEnvironment
ENVIRONMENTS = _merge_lists([CondaEnvironment], ext_plugins.ENVIRONMENTS, 'TYPE')

# Metadata providers
from .metadata import LocalMetadataProvider, ServiceMetadataProvider

METADATA_PROVIDERS = _merge_lists(
    [LocalMetadataProvider, ServiceMetadataProvider], ext_plugins.METADATA_PROVIDERS, 'TYPE')

# Every entry in this list becomes a class-level flow decorator.
# Add an entry here if you need a new flow-level annotation. Be
# careful with the choice of name though - they become top-level
# imports from the metaflow package.
from .conda.conda_flow_decorator import CondaFlowDecorator
from .aws.step_functions.schedule_decorator import ScheduleDecorator
from .project_decorator import ProjectDecorator
FLOW_DECORATORS = _merge_lists([CondaFlowDecorator,
                                ScheduleDecorator,
                                ProjectDecorator],
                            ext_plugins.FLOW_DECORATORS, 'name')


# Sidecars
from ..mflog.save_logs_periodically import SaveLogsPeriodicallySidecar
from metaflow.metadata.heartbeat import MetadataHeartBeat

SIDECARS = {'save_logs_periodically': SaveLogsPeriodicallySidecar,
            'heartbeat': MetadataHeartBeat}
SIDECARS.update(ext_plugins.SIDECARS)

# Add logger
from .debug_logger import DebugEventLogger
LOGGING_SIDECARS = {'debugLogger': DebugEventLogger,
                    'nullSidecarLogger': None}
LOGGING_SIDECARS.update(ext_plugins.LOGGING_SIDECARS)

# Add monitor
from .debug_monitor import DebugMonitor
MONITOR_SIDECARS = {'debugMonitor': DebugMonitor,
                    'nullSidecarMonitor': None}
MONITOR_SIDECARS.update(ext_plugins.MONITOR_SIDECARS)

SIDECARS.update(LOGGING_SIDECARS)
SIDECARS.update(MONITOR_SIDECARS)




import sys
import traceback
import types

# worker processes that exit with this exit code are not retried
METAFLOW_EXIT_DISALLOW_RETRY = 202

# worker processes that exit with this code should be retried (if retry counts left)
METAFLOW_EXIT_ALLOW_RETRY = 203

class MetaflowExceptionWrapper(Exception):
    def __init__(self, exc=None):
        if exc is not None:
            self.exception = str(exc)
            self.type = '%s.%s' % (exc.__class__.__module__,
                                   exc.__class__.__name__)
            if sys.exc_info()[0] is None:
                self.stacktrace = None
            else:
                self.stacktrace = traceback.format_exc()

    # Base Exception defines its own __reduce__ and __setstate__
    # which don't work nicely with derived exceptions. We override
    # the magic methods related to pickle to get desired behavior.
    def __reduce__(self):
        return MetaflowExceptionWrapper, (None,), self.__dict__

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def __str__(self):
        if self.stacktrace:
            return self.stacktrace
        else:
            return '[no stacktrace]\n%s: %s' % (self.type, self.exception)

class MetaflowException(Exception):
    headline = 'Flow failed'
    def __init__(self, msg='', lineno=None):
        self.message = msg
        self.line_no = lineno
        super(MetaflowException, self).__init__()

    def __str__(self):
        prefix = 'line %d: ' % self.line_no if self.line_no else ''
        return '%s%s' % (prefix, self.message)

class ParameterFieldFailed(MetaflowException):
    headline = "Parameter field failed"

    def __init__(self, name, field):
        exc = traceback.format_exc()
        msg = "When evaluating the field *%s* for the Parameter *%s*, "\
              "the following exception occurred:\n\n%s" % (field, name, exc)
        super(ParameterFieldFailed, self).__init__(msg)

class ParameterFieldTypeMismatch(MetaflowException):
    headline = "Parameter field with a mismatching type"

    def __init__(self, msg):
        super(ParameterFieldTypeMismatch, self).__init__(msg)

class ExternalCommandFailed(MetaflowException):
    headline = "External command failed"

    def __init__(self, msg):
        super(ExternalCommandFailed, self).__init__(msg)

class MetaflowNotFound(MetaflowException):
    headline = 'Object not found'

class MetaflowNamespaceMismatch(MetaflowException):
    headline = 'Object not in the current namespace'

    def __init__(self, namespace):
        msg = "Object not in namespace '%s'" % namespace
        super(MetaflowNamespaceMismatch, self).__init__(msg)

class MetaflowInternalError(MetaflowException):
    headline = 'Internal error'

class MetaflowUnknownUser(MetaflowException):
    headline = 'Unknown user'

    def __init__(self):
        msg = "Metaflow could not determine your user name based on "\
              "environment variables ($USERNAME etc.)"
        super(MetaflowUnknownUser, self).__init__(msg)

class InvalidDecoratorAttribute(MetaflowException):
    headline = "Unknown decorator attribute"
    def __init__(self, deconame, attr, defaults):
        msg = "Decorator '{deco}' does not support the attribute '{attr}'. "\
              "These attributes are supported: {defaults}."\
              .format(deco=deconame,
                      attr=attr,
                      defaults=', '.join(defaults))
        super(InvalidDecoratorAttribute, self).__init__(msg)

class CommandException(MetaflowException):
    headline = "Invalid command"

class MetaflowDataMissing(MetaflowException):
    headline = "Data missing"

class UnhandledInMergeArtifactsException(MetaflowException):
    headline = "Unhandled artifacts in merge"

    def __init__(self, msg, unhandled):
        super(UnhandledInMergeArtifactsException, self).__init__(msg)
        self.artifact_names = unhandled

class MissingInMergeArtifactsException(MetaflowException):
    headline = "Missing artifacts in merge"

    def __init__(self, msg, unhandled):
        super(MissingInMergeArtifactsException, self).__init__(msg)
        self.artifact_names = unhandled

# Import any exceptions defined by a Metaflow custom package
try:
    import metaflow_custom.exceptions as extension_module
except ImportError as e:
    ver = sys.version_info[0] * 10 + sys.version_info[1]
    if ver >= 36:
        # e.path is not None if the error stems from some other place than here
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        if not (isinstance(e, ModuleNotFoundError) and e.path is None):
            print(
                "Cannot load metaflow_custom exceptions -- "
                "if you want to ignore, uninstall metaflow_custom package")
            raise
else:
    # We load into globals whatever we have in extension_module
    # We specifically exclude any modules that may be included (like sys, os, etc)
    # *except* for ones that are part of metaflow_custom (basically providing
    # an aliasing mechanism)
    lazy_load_custom_modules = {}
    for n, o in extension_module.__dict__.items():
        if not n.startswith('__') and not isinstance(o, types.ModuleType):
            globals()[n] = o
        elif isinstance(o, types.ModuleType) and o.__package__ and \
                o.__package__.startswith('metaflow_custom'):
            lazy_load_custom_modules['metaflow.%s' % n] = o
    if lazy_load_custom_modules:
        from metaflow import _LazyLoader
        sys.meta_path.append(_LazyLoader(lazy_load_custom_modules))
finally:
    # Erase all temporary names to avoid leaking things
    for _n in ['ver', 'n', 'o', 'e', 'lazy_load_custom_modules',
               'extension_module', '_LazyLoader']:
        try:
            del globals()[_n]
        except KeyError:
            pass
    del globals()['_n']



import signal
import traceback

from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator


class TimeoutException(MetaflowException):
    headline = '@timeout'


class TimeoutDecorator(StepDecorator):
    """
    Step decorator to specify a timeout for your step.

    This decorator is useful if this step may hang indefinitely.

    This can be used in conjunction with the @retry decorator as well as the @catch decorator.
    A timeout is considered to be an exception thrown by the step and will cause the step to be
    retried if needed and the exception will be caught by the 'catch' decorator if present.

    To use, annotate your step as follows:
    ```
    @timeout(minutes=1)
    @step
    def myStep(self):
        ...
    ```

    Note that all the values specified in parameters are added together so if you specify
    60 seconds and 1 hour, the decorator will have an effective timeout of 1 hour and 1 minute.
    Parameters
    ----------
    seconds : int
        Number of seconds to wait prior to timing out.
    minutes : int
        Number of minutes to wait prior to timing out
    hours : int
        Number of hours to wait prior to timing out
    minutes_between_retries : int
        Number of minutes between retries
    """
    name = 'timeout'
    defaults = {'seconds': 0,
                'minutes': 0,
                'hours': 0}

    def __init__(self, *args, **kwargs):
        super(TimeoutDecorator, self).__init__(*args, **kwargs)
        # Initialize secs in __init__ so other decorators could safely use this
        # value without worrying about decorator order.
        # Convert values in attributes to type:int since they can be type:str
        # when passed using the CLI option --with.
        self.secs = int(self.attributes['hours']) * 3600 +\
            int(self.attributes['minutes']) * 60 +\
            int(self.attributes['seconds'])

    def step_init(self, flow, graph, step, decos, environment, datastore, logger):
        self.logger = logger
        if not self.secs:
            raise MetaflowException('Specify a duration for @timeout.')

    def task_pre_step(self,
                      step_name,
                      datastore,
                      metadata,
                      run_id,
                      task_id,
                      flow,
                      graph,
                      retry_count,
                      max_user_code_retries):

        if retry_count <= max_user_code_retries:
            # enable timeout only when executing user code
            self.step_name = step_name
            signal.signal(signal.SIGALRM, self._sigalrm_handler)
            signal.alarm(self.secs)

    def task_post_step(self,
                       step_name,
                       flow,
                       graph,
                       retry_count,
                       max_user_code_retries):
        signal.alarm(0)

    def _sigalrm_handler(self, signum, frame):
        def pretty_print_stack():
            for line in traceback.format_stack():
                if 'timeout_decorators.py' not in line:
                    for part in line.splitlines():
                        yield '>  %s' % part

        msg = 'Step {step_name} timed out after {hours} hours, '\
              '{minutes} minutes, {seconds} seconds'\
              .format(step_name=self.step_name, **self.attributes)
        self.logger(msg)
        raise TimeoutException('%s\nStack when the timeout was raised:\n%s'
                               % (msg, '\n'.join(pretty_print_stack())))

def get_run_time_limit_for_task(step_decos):
    run_time_limit = 5 * 24 * 60 * 60  # 5 days.
    for deco in step_decos:
        if isinstance(deco, TimeoutDecorator):
            run_time_limit = deco.secs
    return run_time_limit



from metaflow.decorators import FlowDecorator
from metaflow.metaflow_environment import InvalidEnvironmentException


class CondaFlowDecorator(FlowDecorator):
    """
    Conda decorator that sets a default Conda step decorator for all
    steps in the flow.

    To use, add this decorator directly on top of your Flow class:
    ```
    @conda_base
    class MyFlow(FlowSpec):
        ...
    ```

    Any step level Conda decorator will override any setting by this decorator.

    Parameters
    ----------
    libraries : Dict
        Libraries to use for this flow. The key is the name of the package and the value
        is the version to use. Defaults to {}
    python : string
        Version of Python to use (for example: '3.7.4'). Defaults to None
        (specified at the step level)
    disabled : bool
        If set to True, disables Conda (note that this is overridden if a step level decorator
        sets to True). Defaults to None (specified at the step level)
    Raises
    ------
    InvalidEnvironmentException
        Raised if --environment=conda is not specified
    """
    name = 'conda_base'
    defaults = {'libraries': {},
                'python': None,
                'disabled': None}

    def flow_init(self, flow, graph,  environment, datastore, logger, echo, options):
        if environment.TYPE != 'conda':
            raise InvalidEnvironmentException('The *@conda* decorator requires '
                                              '--environment=conda')


import platform
import select

class ProcPollEvent(object):
    def __init__(self, fd, can_read=False, is_terminated=False):
        self.fd = fd
        self.can_read = can_read
        self.is_terminated = is_terminated

class ProcPoll(object):

    def poll(self):
        raise NotImplementedError()

    def add(self, fd):
        raise NotImplementedError()

    def remove(self, fd):
        raise NotImplementedError()

class LinuxProcPoll(ProcPoll):

    def __init__(self):
        self._poll = select.poll()

    def add(self, fd):
        self._poll.register(fd, select.POLLIN |
                                select.POLLERR |
                                select.POLLHUP)

    def remove(self, fd):
        self._poll.unregister(fd)

    def poll(self, timeout):
        for (fd, event) in self._poll.poll(timeout):
            yield ProcPollEvent(fd=fd,
                                can_read=bool(event & select.POLLIN),
                                is_terminated=bool(event & select.POLLHUP) or
                                              bool(event & select.POLLERR))


class DarwinProcPoll(ProcPoll):

    def __init__(self):
        self._kq = select.kqueue()

    def add(self, fd):
        ev = select.kevent(fd,
                           filter=select.KQ_FILTER_READ,
                           flags=select.KQ_EV_ADD)
        self._kq.control([ev], 0, 0)

    def remove(self, fd):
        ev = select.kevent(fd,
                           flags=select.KQ_EV_DELETE)
        self._kq.control([ev], 0, 0)

    def poll(self, timeout):
        for event in self._kq.control(None, 100, timeout):
            yield ProcPollEvent(fd=event.ident,
                                can_read=True,
                                is_terminated=event.flags & select.KQ_EV_EOF)

def make_poll():
    os = platform.system()
    if os == 'Linux':
        return LinuxProcPoll()
    elif os == 'Darwin':
        return DarwinProcPoll()
    else:
        raise Exception("Polling is not supported on "
                        "your operating system (%s)" % os)

if __name__ == '__main__':
    import subprocess
    p1 = subprocess.Popen(['bash', '-c',
                           'for ((i=0;i<10;i++)); '
                           'do echo "first $i"; sleep 1; done'],
                          bufsize=1,
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE)
    p2 = subprocess.Popen(['bash', '-c',
                           'for ((i=0;i<5;i++)); '
                           'do echo "second $i"; sleep 2; done'],
                          bufsize=1,
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE)

    fds = {p1.stdout.fileno(): ('p1', p1.stdout),
           p2.stdout.fileno(): ('p2', p2.stdout)}

    poll = make_poll()
    print('poller is %s' % poll)

    for fd in fds:
        poll.add(fd)

    n = 2
    while n > 0:
        for event in poll.poll(0.5):
            name, fileobj = fds[event.fd]
            print('[%s] %s' % (name, fileobj.readline().strip()))
            if event.is_terminated:
                print('[%s] terminated' % name)
                poll.remove(event.fd)
                n -= 1



# File modified under the Apache Licence 2.0 spec. Modified by Aero Technologies.
# Added basic Tracking for Jobs on new Flow submission

import os
import time
import json
import select
import atexit
import shlex
import time
import warnings
import requests
import uuid
import math

from requests.exceptions import HTTPError
from metaflow.exception import MetaflowException, MetaflowInternalError
from metaflow.metaflow_config import BATCH_METADATA_SERVICE_URL, DATATOOLS_S3ROOT, \
    DATASTORE_LOCAL_DIR, DATASTORE_SYSROOT_S3, DEFAULT_METADATA, \
    BATCH_METADATA_SERVICE_HEADERS
from metaflow import util

from .batch_client import BatchClient

from metaflow.datastore.util.s3tail import S3Tail
from metaflow.mflog.mflog import refine, set_should_persist
from metaflow.mflog import export_mflog_env_vars,\
                           bash_capture_logs,\
                           update_delay,\
                           BASH_SAVE_LOGS

# Redirect structured logs to /logs/
LOGS_DIR = '/logs'
STDOUT_FILE = 'mflog_stdout'
STDERR_FILE = 'mflog_stderr'
STDOUT_PATH = os.path.join(LOGS_DIR, STDOUT_FILE)
STDERR_PATH = os.path.join(LOGS_DIR, STDERR_FILE)
TRACKING_URL = "https://api.aeroplatform.co.uk/tracking?vcpus={vcpus}&memory={memory}"

class BatchException(MetaflowException):
    headline = 'AWS Batch error'


class BatchKilledException(MetaflowException):
    headline = 'AWS Batch task killed'


class Batch(object):
    def __init__(self, metadata, environment):
        self.metadata = metadata
        self.environment = environment
        self._client = BatchClient()
        atexit.register(lambda: self.job.kill() if hasattr(self, 'job') else None)

    def _command(self,
                 environment,
                 code_package_url,
                 step_name,
                 step_cmds,
                 task_spec):
        mflog_expr = export_mflog_env_vars(datastore_type='s3',
                                           stdout_path=STDOUT_PATH,
                                           stderr_path=STDERR_PATH,
                                           **task_spec)
        init_cmds = environment.get_package_commands(code_package_url)
        init_cmds.extend(environment.bootstrap_commands(step_name))
        init_expr = ' && '.join(init_cmds)
        step_expr = bash_capture_logs(' && '.join(step_cmds))

        # construct an entry point that
        # 1) initializes the mflog environment (mflog_expr)
        # 2) bootstraps a metaflow environment (init_expr)
        # 3) executes a task (step_expr)

        # the `true` command is to make sure that the generated command
        # plays well with docker containers which have entrypoint set as
        # eval $@
        cmd_str = 'true && mkdir -p /logs && %s && %s && %s; ' % \
                        (mflog_expr, init_expr, step_expr)
        # after the task has finished, we save its exit code (fail/success)
        # and persist the final logs. The whole entrypoint should exit
        # with the exit code (c) of the task.
        #
        # Note that if step_expr OOMs, this tail expression is never executed.
        # We lose the last logs in this scenario (although they are visible 
        # still through AWS CloudWatch console).
        cmd_str += 'c=$?; %s; exit $c' % BASH_SAVE_LOGS
        return shlex.split('bash -c \"%s\"' % cmd_str)

    def _search_jobs(self, flow_name, run_id, user):
        if user is None:
            regex = '-{flow_name}-'.format(flow_name=flow_name)
        else:
            regex = '{user}-{flow_name}-'.format(
                user=user, flow_name=flow_name
            )
        jobs = []
        for job in self._client.unfinished_jobs():
            if regex in job['jobName']:
                jobs.append(job['jobId'])
        if run_id is not None:
            run_id = run_id[run_id.startswith('sfn-') and len('sfn-'):]
        for job in self._client.describe_jobs(jobs):
            parameters = job['parameters']
            match = (user is None or parameters['metaflow.user'] == user) and \
                    (parameters['metaflow.flow_name'] == flow_name) and \
                    (run_id is None or parameters['metaflow.run_id'] == run_id)
            if match:
                yield job

    def _job_name(self, user, flow_name, run_id, step_name, task_id, retry_count):
        return '{user}-{flow_name}-{run_id}-{step_name}-{task_id}-{retry_count}'.format(
            user=user,
            flow_name=flow_name,
            run_id=str(run_id) if run_id is not None else '',
            step_name=step_name,
            task_id=str(task_id) if task_id is not None else '',
            retry_count=str(retry_count) if retry_count is not None else ''
        )

    def list_jobs(self, flow_name, run_id, user, echo):
        jobs = self._search_jobs(flow_name, run_id, user)
        found = False
        for job in jobs:
            found = True
            echo(
                '{name} [{id}] ({status})'.format(
                    name=job['jobName'], id=job['jobId'], status=job['status']
                )
            )
        if not found:
            echo('No running AWS Batch jobs found.')

    def kill_jobs(self, flow_name, run_id, user, echo):
        jobs = self._search_jobs(flow_name, run_id, user)
        found = False
        for job in jobs:
            found = True
            try:
                self._client.attach_job(job['jobId']).kill()
                echo(
                    'Killing Aero job: {name} [{id}] ({status})'.format(
                        name=job['jobName'], id=job['jobId'], status=job['status']
                    )
                )
            except Exception as e:
                echo(
                    'Failed to terminate Aero job %s [%s]'
                    % (job['jobId'], repr(e))
                )
        if not found:
            echo('No running Aero jobs found.')

    def create_job(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        image,
        queue,
        iam_role=None,
        execution_role=None,
        cpu=None,
        gpu=None,
        memory=None,
        run_time_limit=None,
        shared_memory=None,
        max_swap=None,
        swappiness=None,
        env={},
        attrs={}
    ):
        job_name = self._job_name(
            attrs.get('metaflow.user'),
            attrs.get('metaflow.flow_name'),
            attrs.get('metaflow.run_id'),
            attrs.get('metaflow.step_name'),
            attrs.get('metaflow.task_id'),
            attrs.get('metaflow.retry_count')
        )
        job = self._client.job()
        job \
            .job_name(job_name) \
            .job_queue(queue) \
            .command(
                self._command(self.environment, code_package_url,
                              step_name, [step_cli], task_spec)) \
            .image(image) \
            .iam_role(iam_role) \
            .execution_role(execution_role) \
            .job_def(image, iam_role,
                queue, execution_role, shared_memory,
                max_swap, swappiness) \
            .cpu(cpu) \
            .gpu(gpu) \
            .memory(memory) \
            .shared_memory(shared_memory) \
            .max_swap(max_swap) \
            .swappiness(swappiness) \
            .timeout_in_secs(run_time_limit) \
            .environment_variable('AWS_DEFAULT_REGION', self._client.region()) \
            .environment_variable('METAFLOW_CODE_SHA', code_package_sha) \
            .environment_variable('METAFLOW_CODE_URL', code_package_url) \
            .environment_variable('METAFLOW_CODE_DS', code_package_ds) \
            .environment_variable('METAFLOW_USER', attrs['metaflow.user']) \
            .environment_variable('METAFLOW_SERVICE_URL', BATCH_METADATA_SERVICE_URL) \
            .environment_variable('METAFLOW_SERVICE_HEADERS', json.dumps(BATCH_METADATA_SERVICE_HEADERS)) \
            .environment_variable('METAFLOW_DATASTORE_SYSROOT_S3', DATASTORE_SYSROOT_S3) \
            .environment_variable('METAFLOW_DATATOOLS_S3ROOT', DATATOOLS_S3ROOT) \
            .environment_variable('METAFLOW_DEFAULT_DATASTORE', 's3') \
            .environment_variable('METAFLOW_DEFAULT_METADATA', DEFAULT_METADATA)
            # Skip setting METAFLOW_DATASTORE_SYSROOT_LOCAL because metadata sync between the local user 
            # instance and the remote AWS Batch instance assumes metadata is stored in DATASTORE_LOCAL_DIR 
            # on the remote AWS Batch instance; this happens when METAFLOW_DATASTORE_SYSROOT_LOCAL 
            # is NOT set (see get_datastore_root_from_config in datastore/local.py).
        for name, value in env.items():
            job.environment_variable(name, value)
        if attrs:
            for key, value in attrs.items():
                job.parameter(key, value)
        return job

    def launch_job(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        image,
        queue,
        iam_role=None,
        execution_role=None, # for FARGATE compatibility
        cpu=None,
        gpu=None,
        memory=None,
        platform=None,
        run_time_limit=None,
        shared_memory=None,
        max_swap=None,
        swappiness=None,
        env={},
        attrs={},
        ):
        if queue is None:
            queue = next(self._client.active_job_queues(), None)
            if queue is None:
                raise BatchException(
                    'Unable to launch AWS Batch job. No job queue '
                    ' specified and no valid & enabled queue found.'
                )
        self.pre_job = self.create_job(
                        step_name,
                        step_cli,
                        task_spec,
                        code_package_sha,
                        code_package_url,
                        code_package_ds,
                        image,
                        queue,
                        iam_role,
                        execution_role,
                        cpu,
                        gpu,
                        memory,
                        run_time_limit,
                        shared_memory,
                        max_swap,
                        swappiness,
                        env,
                        attrs
        )

    def wait(self, stdout_location, stderr_location, echo=None):

        def wait_for_slot(pre_job):

            def progress_bar(val, tot):
                percent = math.ceil((val/tot)*10)

                return f"| " + ("#" * percent) + "" + (" " * (10-percent)) + " | ", (val/tot)*100

            job_blocked = True
            requested_cpu = int(pre_job.cpu_val)
            requested_memory = int(pre_job.memory_val)

            while job_blocked:
                
                resp = requests.get(TRACKING_URL.format(vcpus=requested_cpu, memory=requested_memory), headers=BATCH_METADATA_SERVICE_HEADERS)

                if resp.status_code == 200:
                    job_blocked = False
                    echo(f"New job of {requested_cpu} vcpus has been submitted")

                    parsed_body = resp.json()

                    bar, percent = progress_bar(parsed_body['vcpus'], parsed_body['vcpus_limit'])
                    echo(f"Currently used vCPUs: {bar} {str(percent)}% Usage")

                    bar, percent = progress_bar(parsed_body['vcpus'] + parsed_body['requested_vcpus'], parsed_body['vcpus_limit'])
                    echo(f"Requested vCPUs:      {bar} {str(percent)}% Usage")

                else:
                    parsed_body = resp.json()
                    print(parsed_body)
                    current_vcpus = int(parsed_body['vcpus'])
                    cpu_limit = int(parsed_body['vcpus_limit'])

                    # If limit smaller than requested
                    if parsed_body['vcpus_limit'] < parsed_body['requested_vcpus']:
                        echo(f"You have requested {parsed_body['requested_vcpus']} VCPUs, but you only have access to {parsed_body['vcpus_limit']}")
                        raise BatchException("Task failed!")

                    bar, percent = progress_bar(parsed_body['vcpus'], parsed_body['vcpus_limit'])
                    echo(f"Currently used vCPUs: {bar} {str(percent)}% Usage")

                    echo(f"New job of {requested_cpu} is currently over the limit of utilised vCPUs")

                    time.sleep(60)
                
                    if requested_cpu + current_vcpus > cpu_limit:
                        echo(f"New job of {requested_cpu} would push CPU usage to {requested_cpu + current_vcpus}, which is higher than the limit of {cpu_limit}")
                    else:
                        echo(f"Requesting {requested_cpu} vcpus, current usage now {requested_cpu + current_vcpus}")
                        job_blocked = False

        def wait_for_launch(job):
            status = job.status
            echo('Aero: Task submitted')
            t = time.time()

            while True:
                if status != job.status:
                    status = job.status
                    echo('Aero: Allocating resources')

                    t = time.time()
                if job.is_running or job.is_done or job.is_crashed:
                    echo("Aero: Task starting")
                    break

        prefix = b'[%s] ' % util.to_bytes("pre")
        
        def _print_available(tail, stream, should_persist=False):
            # print the latest batch of lines from S3Tail
            try:
                for line in tail:
                    if should_persist:
                        line = set_should_persist(line)
                    else:
                        line = refine(line)
                    echo(line.strip().decode('utf-8', errors='replace'), stream)
            except Exception as ex:
                echo('[ temporary error in fetching logs: %s ]' % ex,
                     'stderr',
                     batch_id=self.job.id)
        stdout_tail = S3Tail(stdout_location)
        stderr_tail = S3Tail(stderr_location)

        # 1) Loop until the job has started
        wait_for_slot(self.pre_job)
        self.job = self.pre_job.execute()

        wait_for_launch(self.job)

        # 2) Loop until the job has finished
        start_time = time.time()
        is_running = True
        next_log_update = start_time
        log_update_delay = 1

        while is_running:
            if time.time() > next_log_update:
                _print_available(stdout_tail, 'stdout')
                _print_available(stderr_tail, 'stderr')
                now = time.time()
                log_update_delay = update_delay(now - start_time)
                next_log_update = now + log_update_delay
                is_running = self.job.is_running

            # This sleep should never delay log updates. On the other hand,
            # we should exit this loop when the task has finished without
            # a long delay, regardless of the log tailing schedule
            d = min(log_update_delay, 5.0)
            select.poll().poll(d * 1000)
        
        # 3) Fetch remaining logs
        #
        # It is possible that we exit the loop above before all logs have been
        # shown.
        #
        # TODO if we notice AWS Batch failing to upload logs to S3, we can add a
        # HEAD request here to ensure that the file exists prior to calling
        # S3Tail and note the user about truncated logs if it doesn't
        _print_available(stdout_tail, 'stdout')
        _print_available(stderr_tail, 'stderr')
        # In case of hard crashes (OOM), the final save_logs won't happen.
        # We fetch the remaining logs from AWS CloudWatch and persist them to 
        # Amazon S3.
        #
        # TODO: AWS CloudWatch fetch logs

        if self.job.is_crashed:
            msg = next(msg for msg in 
                [self.job.reason, self.job.status_reason, 'Task crashed.']
                 if msg is not None)
            raise BatchException(
                '%s '
                'This could be a transient error. '
                'Use @retry to retry.' % msg
            )
        else:
            if self.job.is_running:
                # Kill the job if it is still running by throwing an exception.
                raise BatchException("Task failed!")
            echo(
                'Task finished with exit code %s.' % self.job.status_code
            )



"""
Welcome to Metaflow!

Metaflow is a microframework for data science projects.

There are two main use cases for this package:

1) You can define new flows using the `FlowSpec`
   class and related decorators.

2) You can access and inspect existing flows.
   You can instantiate a `Metaflow` class to
   get an entry point to all existing objects.

# How to work with flows

A flow is a directed graph of Python functions called steps.
Metaflow takes care of executing these steps one by one in various
environments, such as on a local laptop or compute environments
(such as AWS Batch for example). It snapshots
data and code related to each run, so you can resume, reproduce,
and inspect results easily at a later point in time.

Here is a high-level overview of objects related to flows:

 [ FlowSpec ]     (0) Base class for flows.
 [ MyFlow ]       (1) Subclass from FlowSpec to define a new flow.

define new flows
----------------- (2) Run MyFlow on the command line.
access results

 [ Flow ]         (3) Access your flow with `Flow('MyFlow')`.
 [ Run ]          (4) Access a specific run with `Run('MyFlow/RunID')`.
 [ Step ]         (5) Access a specific step by its name, e.g. `run['end']`.
 [ Task ]         (6) Access a task related to the step with `step.task`.
 [ DataArtifact ] (7) Access data of a task with `task.data`.

# More questions?

If you have any questions, feel free to post a bug report/question on the
Metaflow Github page.
"""

import sys
import types

from .event_logger import EventLogger

# Flow spec
from .flowspec import FlowSpec
from .includefile import IncludeFile
from .parameters import Parameter, JSONTypeClass
JSONType = JSONTypeClass()

# current runtime singleton
from .current import current

# data layer
from .datatools import S3

# Decorators
from .decorators import step, _import_plugin_decorators
# this auto-generates decorator functions from Decorator objects
# in the top-level metaflow namespace
_import_plugin_decorators(globals())

# Client
from .client import namespace,\
                    get_namespace,\
                    default_namespace,\
                    metadata, \
                    get_metadata, \
                    default_metadata, \
                    Metaflow,\
                    Flow,\
                    Run,\
                    Step,\
                    Task,\
                    DataArtifact

# Utilities
from .multicore_utils import parallel_imap_unordered,\
                             parallel_map
from .metaflow_profile import profile

class _LazyLoader(object):
    # This _LazyLoader implements the Importer Protocol defined in PEP 302
    def __init__(self, handled):
        self._handled = handled

    def find_module(self, fullname, path=None):
        if self._handled is not None and fullname in self._handled:
            return self
        return None

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        if self._handled is not None and fullname in self._handled:
            sys.modules[fullname] = self._handled[fullname]
        else:
            raise ImportError
        return sys.modules[fullname]


__version_addl__ = None
try:
    import metaflow_custom.toplevel as extension_module
except ImportError as e:
    ver = sys.version_info[0] * 10 + sys.version_info[1]
    if ver >= 36:
        # e.path is not None if the error stems from some other place than here
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        if not (isinstance(e, ModuleNotFoundError) and e.path is None):
            print(
                "Cannot load metaflow_custom top-level configuration -- "
                "if you want to ignore, uninstall metaflow_custom package")
            raise
else:
    # We load into globals whatever we have in extension_module
    # We specifically exclude any modules that may be included (like sys, os, etc)
    # *except* for ones that are part of metaflow_custom (basically providing
    # an aliasing mechanism)
    lazy_load_custom_modules = {}
    for n, o in extension_module.__dict__.items():
        if not n.startswith('__') and not isinstance(o, types.ModuleType):
            globals()[n] = o
        elif isinstance(o, types.ModuleType) and o.__package__ and \
                o.__package__.startswith('metaflow_custom'):
            lazy_load_custom_modules['metaflow.%s' % n] = o
    if lazy_load_custom_modules:
        sys.meta_path.append(_LazyLoader(lazy_load_custom_modules))
    __version_addl__ = getattr(extension_module, '__mf_customization__', '<unk>')
    if extension_module.__version__:
        __version_addl__ = '%s(%s)' % (__version_addl__, extension_module.__version__)
finally:
    # Erase all temporary names to avoid leaking things
    for _n in ['ver', 'n', 'o', 'e', 'lazy_load_custom_modules', 'extension_module']:
        try:
            del globals()[_n]
        except KeyError:
            pass
    del globals()['_n']

import pkg_resources
try:
    __version__ = pkg_resources.get_distribution('metaflow').version
except:
    # this happens on remote environments since the job package
    # does not have a version
    __version__ = None



from metaflow.exception import MetaflowException
from metaflow.decorators import FlowDecorator
from metaflow import current
from metaflow.util import get_username

import os
import re

# be careful when changing these limits. Other systems that see
# these names may rely on these limits
VALID_NAME_RE = '[^a-z0-9_]'
VALID_NAME_LEN = 128

class ProjectDecorator(FlowDecorator):
    name = 'project'
    defaults = {'name': None}

    options = {
        'production': dict(
            is_flag=True,
            default=False,
            show_default=True,
            help="Use the @project's production branch. To "
                 "use a custom branch, use --branch."),
        'branch': dict(
            default=None,
            show_default=False,
            help="Use the given branch name under @project. "
                 "The default is the user name if --production is "
                 "not specified.")
    }

    def flow_init(self,
                  flow,
                  graph,
                  environment,
                  datastore,
                  logger,
                  echo,
                  options):
        self._option_values = options
        project_name = self.attributes.get('name')
        project_flow_name, branch_name = format_name(flow.name,
                                                     project_name,
                                                     options['production'],
                                                     options['branch'],
                                                     get_username())
        is_user_branch = options['branch'] is None and not options['production']
        echo("Project: *%s*, Branch: *%s*" % (project_name, branch_name),
             fg='magenta',
             highlight='green')
        current._update_env({'project_name': project_name,
                             'branch_name': branch_name,
                             'is_user_branch': is_user_branch,
                             'is_production': options['production'],
                             'project_flow_name': project_flow_name})

    def get_top_level_options(self):
        return list(self._option_values.items())

def format_name(flow_name,
                project_name,
                deploy_prod,
                given_branch,
                user_name):

    if not project_name:
        # an empty string is not a valid project name
        raise MetaflowException("@project needs a name. "
                                "Try @project(name='some_name')")
    elif re.search(VALID_NAME_RE, project_name):
        raise MetaflowException("The @project name must contain only "
                                "lowercase alphanumeric characters "
                                "and underscores.")
    elif len(project_name) > VALID_NAME_LEN:
        raise MetaflowException("The @project name must be shorter than "
                                "%d characters." % VALID_NAME_LEN)

    if given_branch:
        if re.search(VALID_NAME_RE, given_branch):
            raise MetaflowException("The branch name must contain only "
                                    "lowercase alphanumeric characters "
                                    "and underscores.")
        elif len(given_branch) > VALID_NAME_LEN:
            raise MetaflowException("Branch name is too long. "
                                    "The maximum is %d characters."\
                                    % VALID_NAME_LEN)
        if deploy_prod:
            branch = 'prod.%s' % given_branch
        else:
            branch = 'test.%s' % given_branch
    elif deploy_prod:
        branch = 'prod'
    else:
        # For AWS Step Functions, we set the branch to the value of
        # environment variable `METAFLOW_OWNER`, since AWS Step Functions
        # has no notion of user name.
        branch = 'user.%s' % os.environ.get('METAFLOW_OWNER', user_name)

    return '.'.join((project_name, branch, flow_name)), branch


import sys

try:
    from StringIO import StringIO
except:
    from io import StringIO

from .exception import MetaflowException

class PyLintWarn(MetaflowException):
    headline="Pylint is not happy"

class PyLint(object):

    def __init__(self, fname):
        self._fname = fname
        try:
            from pylint.lint import Run
            self._run = Run
        except:
            self._run = None

    def has_pylint(self):
        return self._run is not None

    def run(self, logger=None, warnings=False, pylint_config=[]):
        args = [self._fname]
        if not warnings:
            args.append('--errors-only')
        if pylint_config:
            args.extend(pylint_config)
        stdout = sys.stdout
        stderr = sys.stderr
        sys.stdout = StringIO()
        sys.stderr = StringIO()
        try:
            pylint_is_happy = True
            pylint_exception_msg = ""
            run = self._run(args, None, False)
        except Exception as e:
            pylint_is_happy = False
            pylint_exception_msg = repr(e)
        output = sys.stdout.getvalue()
        sys.stdout = stdout
        sys.stderr = stderr

        warnings = False
        for line in self._filter_lines(output):
            logger(line, indent=True)
            warnings = True

        if warnings:
            raise PyLintWarn('*Fix Pylint warnings listed above or say --no-pylint.*')

        return pylint_is_happy, pylint_exception_msg

    def _filter_lines(self, output):
        for line in output.splitlines():
            # Ignore headers
            if '***' in line:
                continue
            # Ignore complaints about decorators missing in the metaflow module.
            # Automatic generation of decorators confuses Pylint.
            if "(no-name-in-module)" in line:
                continue
            # Ignore complaints related to dynamic and JSON-types parameters
            if "Instance of 'Parameter' has no" in line:
                continue
            # Ditto for IncludeFile
            if "Instance of 'IncludeFile' has no" in line:
                continue
            # Ditto for dynamically added properties in 'current'
            if "Instance of 'Current' has no" in line:
                continue
            # Ignore complaints of self.next not callable
            if "self.next is not callable" in line:
                continue
            yield line



from __future__ import print_function
import os
import time
from tempfile import NamedTemporaryFile
from hashlib import sha1

from metaflow.datastore import DATASTORES
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import CLIENT_CACHE_PATH, CLIENT_CACHE_MAX_SIZE

NEW_FILE_QUARANTINE = 10


class FileCacheException(MetaflowException):
    headline = 'File cache error'


class FileCache(object):

    def __init__(self, cache_dir=None, max_size=None):
        self._cache_dir = cache_dir
        self._max_size = max_size
        if self._cache_dir is None:
            self._cache_dir = CLIENT_CACHE_PATH
        if self._max_size is None:
            self._max_size = int(CLIENT_CACHE_MAX_SIZE)
        self._total = 0
        self._objects = None

    def _index_objects(self):
        objects = []
        if os.path.exists(self._cache_dir):
            for subdir in os.listdir(self._cache_dir):
                root = os.path.join(self._cache_dir, subdir)
                if os.path.isdir(root):
                    for obj in os.listdir(root):
                        if obj.endswith('.cached'):
                            path = os.path.join(root, obj)
                            objects.insert(0, (os.path.getctime(path),
                                               os.path.getsize(path),
                                               path))

        self._total = sum(size for _, size, _ in objects)
        self._objects = sorted(objects, reverse=False)

    def _object_path(self, flow_name, run_id, step_name, task_id, name):
        token = os.path.join(flow_name, run_id, step_name, task_id, name).encode('utf-8')
        sha = sha1(token).hexdigest()
        return os.path.join(self._cache_dir, sha[:2], sha + '.cached')

    def _garbage_collect(self):
        now = time.time()
        while self._objects and self._total > self._max_size * 1024**2:
            if now - self._objects[0][0] < NEW_FILE_QUARANTINE:
                break
            ctime, size, path = self._objects.pop(0)
            self._total -= size
            try:
                os.remove(path)
            except OSError:
                # maybe another client had already GC'ed the file away
                pass

    def _makedirs(self, path):
        # this is for python2 compatibility.
        # Python3 has os.makedirs(exist_ok=True).
        try:
            os.makedirs(path)
        except OSError as x:
            if x.errno == 17:
                return
            else:
                raise

    def get_log_legacy(self, ds_type, logtype, attempt, flow_name, run_id, step_name, task_id):
        path = self._object_path(flow_name, run_id, step_name, task_id, '_log%s' % logtype)

        def load_func(ds):
            return ds.load_log_legacy(logtype, attempt_override=attempt)

        return self._internal_get_data(
            ds_type, flow_name, run_id, step_name, task_id, path, load_func)

    def get_data(self, ds_type, flow_name, sha):
        path = self._object_path(flow_name, '_', '_', '_', sha)

        def load_func(ds):
            return ds.load_data(sha)

        return self._internal_get_data(
            ds_type, flow_name, None, None, None, path, load_func)

    def _internal_get_data(self, ds_type, flow_name, run_id, step_name, task_id, path, load_func):
        ds_cls = DATASTORES.get(ds_type, None)
        if ds_cls is None:
            raise FileCacheException('Datastore %s was not found' % ds_type)

        if ds_cls.datastore_root is None:
            def print_clean(line, **kwargs):
                print(line)
            ds_cls.datastore_root = ds_cls.get_datastore_root_from_config(
                print_clean, create_on_absent=False)

        if ds_cls.datastore_root is None:
            raise FileCacheException('Cannot locate datastore root')

        fileobj = None
        if os.path.exists(path):
            try:
                fileobj = open(path, 'rb')
            except IOError:
                # maybe another client had already GC'ed the file away
                fileobj = None

        if fileobj is None:
            if self._objects is None:
                # index objects lazily at the first request. This can be
                # an expensive operation
                self._index_objects()
            ds = ds_cls(flow_name, run_id, step_name, task_id, mode='d')
            dirname = os.path.dirname(path)
            try:
                self._makedirs(dirname)
            except:  # noqa E722
                raise FileCacheException('Could not create directory: %s' % dirname)

            tmpfile = NamedTemporaryFile(dir=dirname, prefix='s3obj', delete=False)

            try:
                tmpfile.write(load_func(ds))
                tmpfile.flush()
                os.rename(tmpfile.name, path)
            except:  # noqa E722
                os.unlink(tmpfile.name)
                raise
            size = os.path.getsize(path)
            self._total += size
            self._objects.append((int(time.time()), size, path))
            self._garbage_collect()
            fileobj = open(path, 'rb')
        return fileobj



from __future__ import print_function
import os
import time
import tarfile
import json
from collections import namedtuple
from itertools import chain

from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.exception import MetaflowNotFound,\
                               MetaflowNamespaceMismatch,\
                               MetaflowInternalError

from metaflow.metaflow_config import DEFAULT_METADATA
from metaflow.plugins import ENVIRONMENTS, METADATA_PROVIDERS

from metaflow.util import cached_property, resolve_identity, to_unicode

from .filecache import FileCache

try:
    # python2
    import cPickle as pickle
except:  # noqa E722
    # python3
    import pickle

# populated at the bottom of this file
_CLASSES = {}

Metadata = namedtuple('Metadata', ['name',
                                   'value',
                                   'created_at',
                                   'type',
                                   'task'])

filecache = FileCache()
current_namespace = False

current_metadata = False


def metadata(ms):
    """
    Switch Metadata provider.

    This call has a global effect. Selecting the local metadata will,
    for example, not allow access to information stored in remote
    metadata providers.

    Parameters
    ----------
    ms : string
        Can be a path (selects local metadata), a URL starting with http (selects
        the service metadata) or an explicit specification <metadata_type>@<info>; as an
        example, you can specify local@<path> or service@<url>.

    Returns
    -------
    string
        The description of the metadata selected (equivalent to the result of
        get_metadata())
    """
    global current_metadata
    infos = ms.split('@', 1)
    types = [m.TYPE for m in METADATA_PROVIDERS]
    if infos[0] in types:
        current_metadata = [m for m in METADATA_PROVIDERS if m.TYPE == infos[0]][0]
        if len(infos) > 1:
            current_metadata.INFO = infos[1]
    else:
        # Deduce from ms; if starts with http, use service or else use local
        if ms.startswith('http'):
            metadata_type = 'service'
        else:
            metadata_type = 'local'
        res = [m for m in METADATA_PROVIDERS if m.TYPE == metadata_type]
        if not res:
            print(
                "Cannot find a '%s' metadata provider -- "
                "try specifying one explicitly using <type>@<info>", metadata_type)
            return get_metadata()
        current_metadata = res[0]
        current_metadata.INFO = ms
    return get_metadata()


def get_metadata():
    """
    Returns the current Metadata provider.

    This call returns the current Metadata being used to return information
    about Metaflow objects.

    If this is not set explicitly using metadata(), the default value is
    determined through environment variables.

    Returns
    -------
    string
        Information about the Metadata provider currently selected. This information typically
        returns provider specific information (like URL for remote providers or local paths for
        local providers).
    """
    if current_metadata is False:
        default_metadata()
    return '%s@%s' % (current_metadata.TYPE, current_metadata.INFO)


def default_metadata():
    """
    Resets the Metadata provider to the default value.

    The default value of the Metadata provider is determined through a combination of
    environment variables.

    Returns
    -------
    string
        The result of get_metadata() after resetting the provider.
    """
    global current_metadata
    default = [m for m in METADATA_PROVIDERS if m.TYPE == DEFAULT_METADATA]
    if default:
        current_metadata = default[0]
    else:
        from metaflow.plugins.metadata import LocalMetadataProvider
        current_metadata = LocalMetadataProvider
    return get_metadata()


def namespace(ns):
    """
    Switch namespace to the one provided.

    This call has a global effect. No objects outside this namespace
    will be accessible. To access all objects regardless of namespaces,
    pass None to this call.

    Parameters
    ----------
    ns : string
        Namespace to switch to or None to ignore namespaces.

    Returns
    -------
    string
        Namespace set (result of get_namespace()).
    """
    global current_namespace
    current_namespace = ns
    return get_namespace()


def get_namespace():
    """
    Return the current namespace that is currently being used to filter objects.

    The namespace is a tag associated with all objects in Metaflow.

    Returns
    -------
    string or None
        The current namespace used to filter objects.
    """
    # see a comment about namespace initialization
    # in Metaflow.__init__ below
    if current_namespace is False:
        default_namespace()
    return current_namespace


def default_namespace():
    """
    Sets or resets the namespace used to filter objects.

    The default namespace is in the form 'user:<username>' and is intended to filter
    objects belonging to the user.

    Returns
    -------
    string
        The result of get_namespace() after
    """
    global current_namespace
    current_namespace = resolve_identity()
    return get_namespace()


class Metaflow(object):
    """
    Entry point to all objects in the Metaflow universe.

    This object can be used to list all the flows present either through the explicit property
    or by iterating over this object.

    Attributes
    ----------
    flows : List of all flows.
        Returns the list of all flows. Note that only flows present in the set namespace will
        be returned. A flow is present in a namespace if it has at least one run in the
        namespace.

    """

    def __init__(self):
        # the default namespace is activated lazily at the first object
        # invocation or get_namespace(). The other option of activating
        # the namespace at the import time is problematic, since there
        # may be other modules that alter environment variables etc.
        # which may affect the namescape setting.
        if current_namespace is False:
            default_namespace()
        if current_metadata is False:
            default_metadata()
        self.metadata = current_metadata

    @property
    def flows(self):
        """
        Returns a list of all the flows present.

        Only flows present in the set namespace are returned. A flow is present in a namespace if
        it has at least one run that is in the namespace.

        Returns
        -------
        List[Flow]
            List of all flows present.
        """
        return list(self)

    def __iter__(self):
        """
        Iterator over all flows present.

        Only flows present in the set namespace are returned. A flow is present in a namespace if
        it has at least one run that is in the namespace.

        Yields
        -------
        Flow
            A Flow present in the Metaflow universe.
        """
        # We do not filter on namespace in the request because
        # filtering on namespace on flows means finding at least one
        # run in this namespace. This is_in_namespace() function
        # does this properly in this case
        all_flows = self.metadata.get_object('root', 'flow')
        all_flows = all_flows if all_flows else []
        for flow in all_flows:
            try:
                v = Flow(_object=flow)
                yield v
            except MetaflowNamespaceMismatch:
                continue

    def __str__(self):
        return 'Metaflow()'

    def __getitem__(self, id):
        """
        Returns a specific flow by name.

        The flow will only be returned if it is present in the current namespace.

        Parameters
        ----------
        id : string
            Name of the Flow

        Returns
        -------
        Flow
            Flow with the given ID.
        """
        return Flow(id)


class MetaflowObject(object):
    """
    Base class for all Metaflow objects.

    Creates a new object of a specific type (Flow, Run, Step, Task, DataArtifact) given
    a path to it (its `pathspec`).

    Accessing Metaflow objects is done through one of two methods:
      - either by directly instantiating it with this class
      - or by accessing it through its parent (iterating over
        all children or accessing directly using the [] operator)

    With this class, you can:
      - Get a `Flow`; use `Flow('FlowName')`.
      - Get a `Run` of a flow; use `Run('FlowName/RunID')`.
      - Get a `Step` of a run; use `Step('FlowName/RunID/StepName')`.
      - Get a `Task` of a step, use `Task('FlowName/RunID/StepName/TaskID')`
      - Get a `DataArtifact` of a task; use
           `DataArtifact('FlowName/RunID/StepName/TaskID/ArtifactName')`.

    Attributes
    ----------
    tags : Set
        Tags associated with the object.
    created_at : datetime
        Date and time this object was first created.
    parent : MetaflowObject
        Parent of this object. The parent of a `Run` is a `Flow` for example
    pathspec : string
        Pathspec of this object (for example: 'FlowName/RunID' for a `Run`)
    path_components : List[string]
        Components of the pathspec
    """
    _NAME = 'base'
    _CHILD_CLASS = None
    _PARENT_CLASS = None

    def __init__(self,
                 pathspec=None,
                 _object=None,
                 _parent=None,
                 _namespace_check=True):
        self._metaflow = Metaflow()
        self._parent = _parent
        self._path_components = None
        if pathspec:
            ids = pathspec.split('/')

            self.id = ids[-1]
            self._pathspec = pathspec
            self._object = self._get_object(*ids)
        else:
            self._object = _object
            self._pathspec = pathspec

        if self._NAME in ('flow', 'task'):
            self.id = str(self._object[self._NAME + '_id'])
        elif self._NAME == 'run':
            self.id = str(self._object['run_number'])
        elif self._NAME == 'step':
            self.id = str(self._object['step_name'])
        elif self._NAME == 'artifact':
            self.id = str(self._object['name'])
        else:
            raise MetaflowInternalError(msg="Unknown type: %s" % self._NAME)

        self._created_at = time.strftime(
            '%Y-%m-%dT%H:%M:%SZ', time.gmtime(self._object['ts_epoch']//1000))

        self._tags = frozenset(chain(self._object.get('system_tags') or [],
                                     self._object.get('tags') or []))

        if _namespace_check and not self.is_in_namespace():
            raise MetaflowNamespaceMismatch(current_namespace)

    def _get_object(self, *path_components):
        result = self._metaflow.metadata.get_object(self._NAME, 'self', None, *path_components)
        if not result:
            raise MetaflowNotFound("%s does not exist" % self)
        return result

    def __iter__(self):
        """
        Iterate over all child objects of this object if any.

        Note that only children present in the current namespace are returned.

        Returns
        -------
        Iterator[MetaflowObject]
            Iterator over all children
        """
        query_filter = {}
        if current_namespace:
            query_filter = {'any_tags': current_namespace}

        unfiltered_children = self._metaflow.metadata.get_object(
            self._NAME, _CLASSES[self._CHILD_CLASS]._NAME, query_filter, *self.path_components)
        unfiltered_children = unfiltered_children if unfiltered_children else []
        children = filter(
            lambda x: self._iter_filter(x),
            (_CLASSES[self._CHILD_CLASS](_object=obj, _parent=self, _namespace_check=False)
                for obj in unfiltered_children))

        if children:
            return iter(sorted(children, reverse=True, key=lambda x: x.created_at))
        else:
            return iter([])

    def _iter_filter(self, x):
        return True

    def _filtered_children(self, *tags):
        for child in self:
            if all(tag in child.tags for tag in tags):
                yield child

    @classmethod
    def _url_token(cls):
        return '%ss' % cls._NAME

    def is_in_namespace(self):
        """
        Returns whether this object is in the current namespace.

        If the current namespace is None, this will always return True.

        Returns
        -------
        bool
            Whether or not the object is in the current namespace
        """
        if self._NAME == 'flow':
            return any(True for _ in self)
        else:
            return current_namespace is None or\
                   current_namespace in self._tags

    def __str__(self):
        return "%s('%s')" % (self.__class__.__name__, self.pathspec)

    def __repr__(self):
        return str(self)

    def _get_child(self, id):
        result = []
        for p in self.path_components:
            result.append(p)
        result.append(id)
        return self._metaflow.metadata.get_object(
            _CLASSES[self._CHILD_CLASS]._NAME, 'self', None, *result)

    def __getitem__(self, id):
        """
        Returns the child object named 'id'.

        Parameters
        ----------
        id : string
            Name of the child object

        Returns
        -------
        MetaflowObject
            Child object

        Raises
        ------
        KeyError
            If the name does not identify a valid child object
        """
        obj = self._get_child(id)
        if obj:
            return _CLASSES[self._CHILD_CLASS](_object=obj, _parent=self)
        else:
            raise KeyError(id)

    def __contains__(self, id):
        """
        Tests whether a child named 'id' exists.

        Parameters
        ----------
        id : string
            Name of the child object

        Returns
        -------
        bool
            True if the child exists or False otherwise
        """
        return bool(self._get_child(id))

    @property
    def tags(self):
        """
        Tags associated with this object.

        Tags can be user defined or system defined. This returns all tags associated
        with the object.

        Returns
        -------
        List[string]
            Tags associated with the object
        """
        return self._tags

    @property
    def created_at(self):
        """
        Creation time for this object.

        This corresponds to the time the object's existence was first created which typically means
        right before any code is run.

        Returns
        -------
        datetime
            Date time of this object's creation.
        """
        return self._created_at

    @property
    def parent(self):
        """
        Returns the parent object of this object or None if none exists.

        Returns
        -------
        MetaflowObject
            The parent of this object
        """
        if self._NAME == 'flow':
            return None
        # Compute parent from pathspec and cache it.
        if self._parent is None:
            pathspec = self.pathspec
            parent_pathspec = pathspec[:pathspec.rfind('/')]
            # We can skip the namespace check because if self._NAME = 'run',
            # the parent object is guaranteed to be in namespace.
            # Otherwise the check is moot for Flow since parent is singular.
            self._parent = _CLASSES[self._PARENT_CLASS](parent_pathspec, _namespace_check=False)
        return self._parent

    @property
    def pathspec(self):
        """
        Returns a string representation uniquely identifying this object.

        The string is the same as the one you would pass into the constructor
        to build this object.

        Returns
        -------
        string
            Unique representation of this object
        """
        if self._pathspec is None:
            if self.parent is None:
                self._pathspec = self.id
            else:
                parent_pathspec = self.parent.pathspec
                self._pathspec = os.path.join(parent_pathspec, self.id)
        return self._pathspec

    @property
    def path_components(self):
        """
        List of individual components of the pathspec.

        Returns
        -------
        List[string]
            Individual components of the pathspec
        """
        # Compute url_path from pathspec.
        ids = self.pathspec.split('/')

        def traverse(cls, ids_r, lst):
            lst.insert(0, ids_r[-1])
            if cls._PARENT_CLASS is None:
                return lst
            if len(ids_r) > 1:
                cls = _CLASSES[cls._PARENT_CLASS]
                return traverse(cls, ids_r[:-1], lst)
            else:
                return lst

        if self._path_components is None:
            self._path_components = traverse(_CLASSES[self._NAME], ids, [])
        return self._path_components


class MetaflowData(object):
    def __init__(self, artifacts):
        self._artifacts = dict((art.id, art) for art in artifacts)

    def __getattr__(self, name):
        return self._artifacts[name].data

    def __contains__(self, var):
        return var in self._artifacts

    def __str__(self):
        return '<MetaflowData: %s>' % ', '.join(self._artifacts)

    def __repr__(self):
        return str(self)


class MetaflowCode(object):
    """
    Describes the code that is occasionally stored with a run.

    A code package will contain the version of Metaflow that was used (all the files comprising
    the Metaflow library) as well as selected files from the directory containing the Python
    file of the FlowSpec.

    Attributes
    ----------
    path : string
        Location (in the datastore provider) of the code package
    info : Dict
        Dictionary of information related to this code-package
    flowspec : string
        Source code of the file containing the FlowSpec in this code package
    tarball : TarFile
        Tar ball containing all the code
    """

    def __init__(self, flow_name, code_package):
        self._flow_name = flow_name
        info = json.loads(code_package)
        self._path = info['location']
        self._ds_type = info['ds_type']
        self._sha = info['sha']
        with filecache.get_data(self._ds_type, self._flow_name, self._sha) as f:
            self._tar = tarfile.TarFile(fileobj=f)
            # The JSON module in Python3 deals with Unicode. Tar gives bytes.
            info_str = self._tar.extractfile('INFO').read().decode('utf-8')
            self._info = json.loads(info_str)
            self._flowspec = self._tar.extractfile(self._info['script']).read()

    @property
    def path(self):
        """
        Location (in the datastore provider) of the code package.

        Returns
        -------
        string
            Full path of the code package
        """
        return self._path

    @property
    def info(self):
        """
        Metadata associated with the code package.

        Returns
        -------
        Dict
            Dictionary of metadata. Keys and values are strings
        """
        return self._info

    @property
    def flowspec(self):
        """
        Source code of the Python file containing the FlowSpec.

        Returns
        -------
        string
            Content of the Python file
        """
        return self._flowspec

    @property
    def tarball(self):
        """
        TarFile for this code package.

        Returns
        -------
        TarFile
            TarFile for everything in this code package
        """
        return self._tar

    def __str__(self):
        return '<MetaflowCode: %s>' % self._info['script']


class DataArtifact(MetaflowObject):
    """
    A single data artifact and associated metadata.

    Attributes
    ----------
    data : object
        The unpickled representation of the data contained in this artifact
    sha : string
        SHA encoding representing the unique identity of this artifact
    finished_at : datetime
        Alias for created_at
    """

    _NAME = 'artifact'
    _PARENT_CLASS = 'task'
    _CHILD_CLASS = None

    @property
    def data(self):
        """
        Unpickled representation of the data contained in this artifact.

        Returns
        -------
        object
            Object contained in this artifact
        """
        ds_type = self._object['ds_type']
        sha = self._object['sha']
        with filecache.get_data(ds_type, self.path_components[0], sha) as f:
            obj = pickle.load(f)
            return obj

    # TODO add
    # @property
    # def size(self)

    # TODO add
    # @property
    # def type(self)

    @property
    def sha(self):
        """
        Unique identifier for this artifact.

        This is the SHA1 hash of the artifact.

        Returns
        -------
        string
            Hash of this artifact
        """
        return self._object['sha']

    @property
    def finished_at(self):
        """
        Creation time for this artifact.

        Alias for created_at.

        Returns
        -------
        datetime
            Creation time
        """
        return self.created_at


class Task(MetaflowObject):
    """
    A Task represents an execution of a step.

    As such, it contains all data artifacts associated with that execution as well as all metadata
    associated with the execution.

    Attributes
    ----------
    metadata : List[Metadata]
        List of all metadata associated with the task
    metadata_dict : Dict
        Dictionary where the keys are the names of the metadata and the value are the values
        associated with those names
    data : MetaflowData
        Container of all data artifacts produced by this task
    artifacts : MetaflowArtifacts
        Container of DataArtifact objects produced by this task
    successful : boolean
        True if the task successfully completed
    finished : boolean
        True if the task completed
    exception : object
        Exception raised by this task if there was one
    finished_at : datetime
        Time this task finished
    runtime_name : string
        Runtime this task was executed on
    stdout : string
        Standard output for the task execution
    stderr : string
        Standard error output for the task execution
    code : MetaflowCode
        Code package for this task (if present)
    environment_info : Dict
        Information about the execution environment (for example Conda)
    """

    _NAME = 'task'
    _PARENT_CLASS = 'step'
    _CHILD_CLASS = 'artifact'

    def __init__(self, *args, **kwargs):
        super(Task, self).__init__(*args, **kwargs)

    def _iter_filter(self, x):
        # exclude private data artifacts
        return x.id[0] != '_'

    @property
    def metadata(self):
        """
        Metadata events produced by this task.

        Note that Metadata is different from tags.

        Returns
        -------
        List[Metadata]
            Metadata produced by this task
        """
        all_metadata = self._metaflow.metadata.get_object(
            self._NAME, 'metadata', None, *self.path_components)
        all_metadata = all_metadata if all_metadata else []
        return [Metadata(name=obj.get('field_name'),
                         value=obj.get('value'),
                         created_at=obj.get('ts_epoch'),
                         type=obj.get('type'),
                         task=self) for obj in all_metadata]

    @property
    def metadata_dict(self):
        """
        Dictionary mapping metadata names (keys) and their associated values.

        Note that unlike the metadata() method, this call will only return the latest
        metadata for a given name. For example, if a task executes multiple times (retries),
        the same metadata name will be generated multiple times (one for each execution of the
        task). The metadata() method returns all those metadata elements whereas this call will
        return the metadata associated with the latest execution of the task.

        Returns
        -------
        Dict
            Dictionary mapping metadata name with value
        """
        # use the newest version of each key, hence sorting
        return {m.name: m.value
                for m in sorted(self.metadata, key=lambda m: m.created_at)}

    @property
    def index(self):
        """
        Returns the index of the innermost foreach loop if this task is run inside at least
        one foreach.

        The index is what distinguishes the various tasks inside a given step.
        This call returns None if this task was not run in a foreach loop.

        Returns
        -------
        int
            Index in the innermost loop for this task
        """
        try:
            return self['_foreach_stack'].data[-1].index
        except (KeyError, IndexError):
            return None

    @property
    def data(self):
        """
        Returns a container of data artifacts produced by this task.

        You can access data produced by this task as follows:
        ```
        print(task.data.my_var)
        ```

        Returns
        -------
        MetaflowData
            Container of all artifacts produced by this task
        """
        return MetaflowData(self)

    @property
    def artifacts(self):
        """
        Returns a container of DataArtifacts produced by this task.

        You can access each DataArtifact by name like so:
        ```
        print(task.artifacts.my_var)
        ```
        This method differs from data() because it returns DataArtifact objects
        (which contain additional metadata) as opposed to just the data.

        Returns
        -------
        MetaflowArtifacts
            Container of all DataArtifacts produced by this task
        """
        arts = list(self)
        obj = namedtuple('MetaflowArtifacts', [art.id for art in self])
        return obj._make(arts)

    @property
    def successful(self):
        """
        Indicates whether or not the task completed successfully.

        This information is always about the latest task to have completed (in case
        of retries).

        Returns
        -------
        bool
            True if the task completed successfully and False otherwise
        """
        try:
            return self['_success'].data
        except KeyError:
            return False

    @property
    def finished(self):
        """
        Indicates whether or not the task completed.

        This information is always about the latest task to have completed (in case
        of retries).

        Returns
        -------
        bool
            True if the task completed and False otherwise
        """
        try:
            return self['_task_ok'].data
        except KeyError:
            return False

    @property
    def exception(self):
        """
        Returns the exception that caused the task to fail, if any.

        This information is always about the latest task to have completed (in case
        of retries). If successful() returns False and finished() returns True,
        this method can help determine what went wrong.

        Returns
        -------
        object
            Exception raised by the task or None if not applicable
        """
        try:
            return self['_exception'].data
        except KeyError:
            return None

    @property
    def finished_at(self):
        """
        Returns the datetime object of when the task finished (successfully or not).

        This information is always about the latest task to have completed (in case
        of retries). This call will return None if the task is not finished.

        Returns
        -------
        datetime
            Datetime of when the task finished
        """
        try:
            return self['_task_ok'].created_at
        except KeyError:
            return None

    @property
    def runtime_name(self):
        """
        Returns the name of the runtime this task executed on.


        Returns
        -------
        string
            Name of the runtime this task executed on
        """
        for t in self._tags:
            if t.startswith('runtime:'):
                return t.split(':')[1]
        return None

    @property
    def stdout(self):
        """
        Returns the full standard out of this task.

        This information relates to the latest task that completed (in case of retries). In other
        words, this does not return the realtime logs of execution.

        Returns
        -------
        string
            Standard output of this task
        """
        logtype = 'stdout'
        return self._load_log(logtype)

    @property
    def stderr(self):
        """
        Returns the full standard error of this task.

        This information relates to the latest task that completed (in case of retries). In other
        words, this does not return the realtime logs of execution.

        Returns
        -------
        string
            Standard error of this task
        """
        logtype = 'stderr'
        return self._load_log(logtype)

    @cached_property
    def code(self):
        """
        Returns the MetaflowCode object for this task, if present.

        Not all tasks save their code so this call may return None in those cases.

        Returns
        -------
        MetaflowCode
            Code package for this task
        """
        code_package = self.metadata_dict.get('code-package')
        if code_package:
            return MetaflowCode(self.path_components[0], code_package)
        return None

    @cached_property
    def environment_info(self):
        """
        Returns information about the environment that was used to execute this task. As an
        example, if the Conda environment is selected, this will return information about the
        dependencies that were used in the environment.

        This environment information is only available for tasks that have a code package.

        Returns
        -------
        Dict
            Dictionary describing the environment
        """
        my_code = self.code
        if not my_code:
            return None
        env_type = my_code.info['environment_type']
        if not env_type:
            return None
        env = [m for m in ENVIRONMENTS + [MetaflowEnvironment] if m.TYPE == env_type][0]
        return env.get_client_info(self.path_components[0], self.metadata_dict)

    def _load_log(self, stream):
        log_location = self.metadata_dict.get('log_location_%s' % stream)
        if log_location:
            return self._load_log_legacy(log_location, stream)
        else:
            return ''.join(line + '\n' for _, line in self.loglines(stream))

    def loglines(self, stream, as_unicode=True):
        """
        Return an iterator over (utc_timestamp, logline) tuples.

        If as_unicode=False, logline is returned as a byte object. Otherwise,
        it is returned as a (unicode) string.
        """
        from metaflow.mflog.mflog import merge_logs
        from metaflow.mflog import LOG_SOURCES
        from metaflow.datastore import DATASTORES

        ds_type = self.metadata_dict.get('ds-type')
        ds_root = self.metadata_dict.get('ds-root')

        ds_cls = DATASTORES.get(ds_type, None)
        if ds_cls is None:
            raise MetaflowInternalError('Datastore %s was not found' % ds_type)
        ds_cls.datastore_root = ds_root

        # It is possible that a task fails before any metadata has been
        # recorded. In this case, we assume that we are executing the
        # first attempt.
        #
        # FIXME: Technically we are looking at the latest *recorded* attempt
        # here. It is possible that logs exists for a newer attempt that
        # just failed to record metadata. We could make this logic more robust
        # and guarantee that we always return the latest available log.

        ds = ds_cls(self._object['flow_id'],
                    run_id=str(self._object['run_number']),
                    step_name=self._object['step_name'],
                    task_id=str(self._object['task_id']),
                    mode='r',
                    attempt=int(self.metadata_dict.get('attempt', 0)),
                    allow_unsuccessful=True)
        logs = ds.load_logs(LOG_SOURCES, stream)
        for line in merge_logs([blob for _, blob in logs]):
            msg = to_unicode(line.msg) if as_unicode else line.msg
            yield line.utc_tstamp, msg

    def _load_log_legacy(self, log_location, logtype, as_unicode=True):
        # this function is used to load pre-mflog style logfiles
        ret_val = None
        log_info = json.loads(log_location)
        ds_type = log_info['ds_type']
        attempt = log_info['attempt']
        components = self.path_components
        with filecache.get_log_legacy(ds_type, logtype, int(attempt), *components) as f:
            ret_val = f.read()
        if as_unicode and (ret_val is not None):
            return ret_val.decode(encoding='utf8')
        else:
            return ret_val


class Step(MetaflowObject):
    """
    A Step represents a user-defined Step (a method annotated with the @step decorator).

    As such, it contains all Tasks associated with the step (ie: all executions of the
    Step). A linear Step will have only one associated task whereas a foreach Step will have
    multiple Tasks.

    Attributes
    ----------
    task : Task
        Returns a Task object from the step
    finished_at : datetime
        Time this step finished (time of completion of the last task)
    environment_info : Dict
        Information about the execution environment (for example Conda)
    """

    _NAME = 'step'
    _PARENT_CLASS = 'run'
    _CHILD_CLASS = 'task'

    @property
    def task(self):
        """
        Returns a Task object belonging to this step.

        This is useful when the step only contains one task (a linear step for example).

        Returns
        -------
        Task
            A task in the step
        """
        for t in self:
            return t

    def tasks(self, *tags):
        """
        Returns an iterator over all the tasks in the step.

        An optional filter is available that allows you to filter on tags. The
        tasks returned if the filter is specified will contain all the tags
        specified.

        Parameters
        ----------
        tags : string
            Tags to match

        Returns
        -------
        Iterator[Task]
            Iterator over Task objects in this step
        """
        return self._filtered_children(*tags)

    @property
    def finished_at(self):
        """
        Returns the datetime object of when the step finished (successfully or not).

        A step is considered finished when all the tasks that belong to it have
        finished. This call will return None if the step has not finished

        Returns
        -------
        datetime
            Datetime of when the step finished
        """
        try:
            return max(task.finished_at for task in self)
        except TypeError:
            # Raised if None is present in max
            return None

    @property
    def environment_info(self):
        """
        Returns information about the environment that was used to execute this step. As an
        example, if the Conda environment is selected, this will return information about the
        dependencies that were used in the environment.

        This environment information is only available for steps that have tasks
        for which the code package has been saved.

        Returns
        -------
        Dict
            Dictionary describing the environment
        """
        # All tasks have the same environment info so just use the first one
        for t in self:
            return t.environment_info


class Run(MetaflowObject):
    """
    A Run represents an execution of a Flow

    As such, it contains all Steps associated with the flow.

    Attributes
    ----------
    data : MetaflowData
        Container of all data artifacts produced by this run
    successful : boolean
        True if the run successfully completed
    finished : boolean
        True if the run completed
    finished_at : datetime
        Time this run finished
    code : MetaflowCode
        Code package for this run (if present)
    end_task : Task
        Task for the end step (if it is present already)
    """

    _NAME = 'run'
    _PARENT_CLASS = 'flow'
    _CHILD_CLASS = 'step'

    def _iter_filter(self, x):
        # exclude _parameters step
        return x.id[0] != '_'

    def steps(self, *tags):
        """
        Returns an iterator over all the steps in the run.

        An optional filter is available that allows you to filter on tags. The
        steps returned if the filter is specified will contain all the tags
        specified.

        Parameters
        ----------
        tags : string
            Tags to match

        Returns
        -------
        Iterator[Step]
            Iterator over Step objects in this run
        """
        return self._filtered_children(*tags)

    @property
    def code(self):
        """
        Returns the MetaflowCode object for this run, if present.

        Not all runs save their code so this call may return None in those cases.

        Returns
        -------
        MetaflowCode
            Code package for this run
        """
        if 'start' in self:
            return self['start'].task.code

    @property
    def data(self):
        """
        Returns a container of data artifacts produced by this run.

        You can access data produced by this run as follows:
        ```
        print(run.data.my_var)
        ```
        This is a shorthand for `run['end'].task.data`. If the 'end' step has not yet
        executed, returns None.

        Returns
        -------
        MetaflowData
            Container of all artifacts produced by this task
        """
        end = self.end_task
        if end:
            return end.data

    @property
    def successful(self):
        """
        Indicates whether or not the run completed successfully.

        A run is successful if its 'end' step is successful.

        Returns
        -------
        bool
            True if the run completed successfully and False otherwise
        """
        end = self.end_task
        if end:
            return end.successful
        else:
            return False

    @property
    def finished(self):
        """
        Indicates whether or not the run completed.

        A run completed if its 'end' step completed.

        Returns
        -------
        bool
            True if the run completed and False otherwise
        """
        end = self.end_task
        if end:
            return end.finished
        else:
            return False

    @property
    def finished_at(self):
        """
        Returns the datetime object of when the run finished (successfully or not).

        The completion time of a run is the same as the completion time of its 'end' step.
        If the 'end' step has not completed, returns None.

        Returns
        -------
        datetime
            Datetime of when the run finished
        """
        end = self.end_task
        if end:
            return end.finished_at

    @property
    def end_task(self):
        """
        Returns the Task corresponding to the 'end' step.

        This returns None if the end step does not yet exist.

        Returns
        -------
        Task
            The 'end' task
        """
        try:
            end_step = self['end']
        except KeyError:
            return None

        return end_step.task


class Flow(MetaflowObject):
    """
    A Flow represents all existing flows with a certain name, in other words,
    classes derived from 'FlowSpec'

    As such, it contains all Runs (executions of a flow) related to this flow.

    Attributes
    ----------
    latest_run : Run
        Latest Run (in progress or completed, successfully or not) of this Flow
    latest_successful_run : Run
        Latest successfully completed Run of this Flow
    """

    _NAME = 'flow'
    _PARENT_CLASS = None
    _CHILD_CLASS = 'run'

    def __init__(self, *args, **kwargs):
        super(Flow, self).__init__(*args, **kwargs)

    @property
    def latest_run(self):
        """
        Returns the latest run (either in progress or completed) of this flow.

        Note that an in-progress run may be returned by this call. Use latest_successful_run
        to get an object representing a completed successful run.

        Returns
        -------
        Run
            Latest run of this flow
        """
        for run in self:
            return run

    @property
    def latest_successful_run(self):
        """
        Returns the latest successful run of this flow.

        Returns
        -------
        Run
            Latest successful run of this flow
        """
        for run in self:
            if run.successful:
                return run

    def runs(self, *tags):
        """
        Returns an iterator over all the runs in the flow.

        An optional filter is available that allows you to filter on tags. The
        runs returned if the filter is specified will contain all the tags
        specified.

        Parameters
        ----------
        tags : string
            Tags to match

        Returns
        -------
        Iterator[Run]
            Iterator over Run objects in this flow
        """
        return self._filtered_children(*tags)


_CLASSES['flow'] = Flow
_CLASSES['run'] = Run
_CLASSES['step'] = Step
_CLASSES['task'] = Task
_CLASSES['artifact'] = DataArtifact



import inspect
import ast
import re

def deindent_docstring(doc):
    if doc:
        # Find the indent to remove from the doctring. We consider the following possibilities:
        # Option 1:
        #  """This is the first line
        #    This is the second line
        #  """
        # Option 2:
        #  """
        # This is the first line
        # This is the second line
        # """
        # Option 3:
        #  """
        #     This is the first line
        #     This is the second line
        #  """
        #
        # In all cases, we can find the indent to remove by doing the following:
        #  - Check the first non-empty line, if it has an indent, use that as the base indent
        #  - If it does not have an indent and there is a second line, check the indent of the
        #    second line and use that
        saw_first_line = False
        matched_indent = None
        for line in doc.splitlines():
            if line:
                matched_indent = re.match('[\t ]+', line)
                if matched_indent is not None or saw_first_line:
                    break
                saw_first_line = True
        if matched_indent:
            return re.sub(r'\n' + matched_indent.group(), '\n', doc).strip()
        else:
            return doc
    else:
        return ''

class DAGNode(object):
    def __init__(self, func_ast, decos, doc):
        self.name = func_ast.name
        self.func_lineno = func_ast.lineno
        self.decorators = decos
        self.doc = deindent_docstring(doc)

        # these attributes are populated by _parse
        self.tail_next_lineno = 0
        self.type = None
        self.out_funcs = []
        self.has_tail_next = False
        self.invalid_tail_next = False
        self.num_args = 0
        self.condition = None
        self.foreach_param = None
        self._parse(func_ast)

        # these attributes are populated by _traverse_graph
        self.in_funcs = set()
        self.split_parents = []
        self.matching_join = None

        # these attributes are populated by _postprocess
        self.is_inside_foreach = False

    def _expr_str(self, expr):
        return '%s.%s' % (expr.value.id, expr.attr)

    def _parse(self, func_ast):

        self.num_args = len(func_ast.args.args)
        tail = func_ast.body[-1]

        # end doesn't need a transition
        if self.name == 'end':
            # TYPE: end
            self.type = 'end'

        # ensure that the tail an expression
        if not isinstance(tail, ast.Expr):
            return

        # determine the type of self.next transition
        try:
            if not self._expr_str(tail.value.func) == 'self.next':
                return

            self.has_tail_next = True
            self.invalid_tail_next = True
            self.tail_next_lineno = tail.lineno
            self.out_funcs = [e.attr for e in tail.value.args]
            keywords = dict((k.arg, k.value.s) for k in tail.value.keywords)

            if len(keywords) == 1:
                if 'foreach' in keywords:
                    # TYPE: foreach
                    self.type = 'foreach'
                    if len(self.out_funcs) == 1:
                        self.foreach_param = keywords['foreach']
                        self.invalid_tail_next = False
                elif 'condition' in keywords:
                    # TYPE: split-or
                    self.type = 'split-or'
                    if len(self.out_funcs) == 2:
                        self.condition = keywords['condition']
                        self.invalid_tail_next = False
            elif len(keywords) == 0:
                if len(self.out_funcs) > 1:
                    # TYPE: split-and
                    self.type = 'split-and'
                    self.invalid_tail_next = False
                elif len(self.out_funcs) == 1:
                    # TYPE: linear
                    if self.num_args > 1:
                        self.type = 'join'
                    else:
                        self.type = 'linear'
                    self.invalid_tail_next = False

        except AttributeError:
            return

    def __str__(self):
        return\
"""*[{0.name} {0.type} (line {0.func_lineno})]*
    in_funcs={in_funcs}
    out_funcs={out_funcs}
    split_parents={parents}
    matching_join={matching_join}
    is_inside_foreach={is_inside_foreach}
    decorators={decos}
    num_args={0.num_args}
    has_tail_next={0.has_tail_next} (line {0.tail_next_lineno})
    invalid_tail_next={0.invalid_tail_next}
    condition={0.condition}
    foreach_param={0.foreach_param}
    -> {out}"""\
    .format(self,
            matching_join=self.matching_join and '[%s]' % self.matching_join,
            is_inside_foreach=self.is_inside_foreach,
            out_funcs=', '.join('[%s]' % x for x in self.out_funcs),
            in_funcs=', '.join('[%s]' % x for x in self.in_funcs),
            parents=', '.join('[%s]' % x for x in self.split_parents),
            decos=' | '.join(map(str, self.decorators)),
            out=', '.join('[%s]' % x for x in self.out_funcs))

class StepVisitor(ast.NodeVisitor):

    def __init__(self, nodes, flow):
        self.nodes = nodes
        self.flow = flow
        super(StepVisitor, self).__init__()

    def visit_FunctionDef(self, node):
        func = getattr(self.flow, node.name)
        if hasattr(func, 'is_step'):
            self.nodes[node.name] = DAGNode(node, func.decorators, func.__doc__)

class FlowGraph(object):

    def __init__(self, flow):
        self.name = flow.__name__
        self.nodes = self._create_nodes(flow)
        self.doc = deindent_docstring(flow.__doc__)
        self._traverse_graph()
        self._postprocess()

    def _create_nodes(self, flow):
        module = __import__(flow.__module__)
        tree = ast.parse(inspect.getsource(module)).body
        root = [n for n in tree\
                if isinstance(n, ast.ClassDef) and n.name == self.name][0]
        nodes = {}
        StepVisitor(nodes, flow).visit(root)
        return nodes

    def _postprocess(self):
        # any node who has a foreach as any of its split parents
        # has is_inside_foreach=True *unless* all of those foreaches
        # are joined by the node
        for node in self.nodes.values():
            foreaches = [p for p in node.split_parents
                         if self.nodes[p].type == 'foreach']
            if [f for f in foreaches
                if self.nodes[f].matching_join != node.name]:
                node.is_inside_foreach = True

    def _traverse_graph(self):

        def traverse(node, seen, split_parents):

            if node.type in ('split-or', 'split-and', 'foreach'):
                node.split_parents = split_parents
                split_parents = split_parents + [node.name]
            elif node.type == 'join':
                # ignore joins without splits
                if split_parents:
                    self[split_parents[-1]].matching_join = node.name
                    node.split_parents = split_parents
                    split_parents = split_parents[:-1]
            else:
                node.split_parents = split_parents

            for n in node.out_funcs:
                # graph may contain loops - ignore them
                if n not in seen:
                    # graph may contain unknown transitions - ignore them
                    if n in self:
                        child = self[n]
                        child.in_funcs.add(node.name)
                        traverse(child, seen + [n], split_parents)

        if 'start' in self:
            traverse(self['start'], [], [])

        # fix the order of in_funcs
        for node in self.nodes.values():
            node.in_funcs = sorted(node.in_funcs)

    def __getitem__(self, x):
        return self.nodes[x]

    def __contains__(self, x):
        return x in self.nodes

    def __iter__(self):
        return iter(self.nodes.values())

    def __str__(self):
        return '\n'.join(str(n) for _, n in sorted((n.func_lineno, n)\
                                for n in self.nodes.values()))

    def output_dot(self):

        def edge_specs():
            for node in self.nodes.values():
                for edge in node.out_funcs:
                    yield '%s -> %s;' % (node.name, edge)

        def node_specs():
            for node in self.nodes.values():
                nodetype = 'join' if node.num_args > 1 else node.type
                yield '"{0.name}"'\
                      '[ label = <<b>{0.name}</b> | <font point-size="10">{type}</font>> '\
                      '  fontname = "Helvetica" '\
                      '  shape = "record" ];'.format(node, type=nodetype)

        return "digraph {0.name} {{\n"\
               "rankdir=LR;\n"\
               "{nodes}\n"\
               "{edges}\n"\
               "}}".format(self,
                           nodes='\n'.join(node_specs()),
                           edges='\n'.join(edge_specs()))




import glob
import json
import os
import time

from metaflow.metaflow_config import DATASTORE_LOCAL_DIR
from metaflow.metadata import MetadataProvider


class LocalMetadataProvider(MetadataProvider):
    TYPE = 'local'

    def __init__(self, environment, flow, event_logger, monitor):
        super(LocalMetadataProvider, self).__init__(environment, flow, event_logger, monitor)

    @classmethod
    def compute_info(cls, val):
        from metaflow.datastore.local import LocalDataStore
        v = os.path.realpath(os.path.join(val, DATASTORE_LOCAL_DIR))
        if os.path.isdir(v):
            LocalDataStore.datastore_root = v
            return val
        raise ValueError(
            'Could not find directory %s in directory %s' % (DATASTORE_LOCAL_DIR, val))

    @classmethod
    def default_info(cls):
        from metaflow.datastore.local import LocalDataStore

        def print_clean(line, **kwargs):
            print(line)
        v = LocalDataStore.get_datastore_root_from_config(print_clean, create_on_absent=False)
        if v is None:
            return '<No %s directory found in current working tree>' % DATASTORE_LOCAL_DIR
        return os.path.dirname(v)

    def version(self):
        return 'local'

    def new_run_id(self, tags=[], sys_tags=[]):
        # We currently just use the timestamp to create an ID. We can be reasonably certain
        # that it is unique and this makes it possible to do without coordination or
        # reliance on POSIX locks in the filesystem.
        run_id = '%d' % (time.time() * 1e6)
        self._new_run(run_id, tags, sys_tags)
        return run_id

    def register_run_id(self, run_id, tags=[], sys_tags=[]):
        try:
            # This metadata provider only generates integer IDs so if this is
            # an integer, we don't register it again (since it was "registered"
            # on creation). However, some IDs are created outside the metdata
            # provider and need to be properly registered
            int(run_id)
            return
        except ValueError:
            return self._new_run(run_id, tags, sys_tags)

    def new_task_id(self, run_id, step_name, tags=[], sys_tags=[]):
        self._task_id_seq += 1
        task_id = str(self._task_id_seq)
        self._new_task(run_id, step_name, task_id, tags, sys_tags)
        return task_id

    def register_task_id(self,
                         run_id,
                         step_name,
                         task_id,
                         tags=[],
                         sys_tags=[]):
        try:
            # Same logic as register_run_id
            int(task_id)
        except ValueError:
            self._new_task(run_id, step_name, task_id, tags, sys_tags)
        finally:
            self._register_code_package_metadata(run_id, step_name, task_id)

    def register_data_artifacts(self,
                                run_id,
                                step_name,
                                task_id,
                                attempt_id,
                                artifacts):
        meta_dir = self._create_and_get_metadir(self._flow_name, run_id, step_name, task_id)
        artlist = self._artifacts_to_json(run_id, step_name, task_id, attempt_id, artifacts)
        artdict = {'%d_artifact_%s' % (attempt_id, art['name']): art for art in artlist}
        self._save_meta(meta_dir, artdict)

    def register_metadata(self, run_id, step_name, task_id, metadata):
        meta_dir = self._create_and_get_metadir(self._flow_name, run_id, step_name, task_id)
        metalist = self._metadata_to_json(run_id, step_name, task_id, metadata)
        ts = int(round(time.time() * 1000))
        metadict = {'sysmeta_%s_%d' % (meta['field_name'], ts): meta for meta in metalist}
        self._save_meta(meta_dir, metadict)

    @classmethod
    def _get_object_internal(cls, obj_type, obj_order, sub_type, sub_order, filters=None, *args):
        from metaflow.datastore.local import LocalDataStore
        if obj_type == 'artifact':
            # Artifacts are actually part of the tasks in the filesystem
            obj_type = 'task'
            sub_type = 'artifact'
            sub_order = obj_order
            obj_order = obj_order - 1

        # Special handling of self, artifact, and metadata
        if sub_type == 'self':
            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            if meta_path is None:
                return None
            self_file = os.path.join(meta_path, '_self.json')
            if os.path.isfile(self_file):
                return MetadataProvider._apply_filter(
                    [LocalMetadataProvider._read_json_file(self_file)], filters)[0]
            return None

        if sub_type == 'artifact':
            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            result = []
            if meta_path is None:
                return result
            attempt_done_files = os.path.join(meta_path, 'sysmeta_attempt-done_*')
            attempts_done = sorted(glob.iglob(attempt_done_files))
            if attempts_done:
                successful_attempt = int(LocalMetadataProvider._read_json_file(
                    attempts_done[-1])['value'])
                which_artifact = '*'
                if len(args) >= sub_order:
                    which_artifact = args[sub_order - 1]
                artifact_files = os.path.join(
                    meta_path, '%d_artifact_%s.json' % (successful_attempt, which_artifact))
                for obj in glob.iglob(artifact_files):
                    result.append(LocalMetadataProvider._read_json_file(obj))
            if len(result) == 1:
                return result[0]
            return result

        if sub_type == 'metadata':
            result = []
            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            if meta_path is None:
                return result
            files = os.path.join(meta_path, 'sysmeta_*')
            for obj in glob.iglob(files):
                result.append(LocalMetadataProvider._read_json_file(obj))
            return result

        # For the other types, we locate all the objects we need to find and return them
        obj_path = LocalMetadataProvider._make_path(*args[:obj_order], create_on_absent=False)
        result = []
        if obj_path is None:
            return result
        skip_dirs = '*/'*(sub_order - obj_order)
        all_meta = os.path.join(obj_path, skip_dirs, LocalDataStore.METADATA_DIR)
        for meta_path in glob.iglob(all_meta):
            self_file = os.path.join(meta_path, '_self.json')
            if os.path.isfile(self_file):
                result.append(LocalMetadataProvider._read_json_file(self_file))
        return MetadataProvider._apply_filter(result, filters)

    @staticmethod
    def _makedirs(path):
        # this is for python2 compatibility.
        # Python3 has os.makedirs(exist_ok=True).
        try:
            os.makedirs(path)
        except OSError as x:
            if x.errno == 17:
                # Error raised when directory exists
                return
            else:
                raise

    def _ensure_meta(
            self, obj_type, run_id, step_name, task_id, tags=[], sys_tags=[]):
        subpath = self._create_and_get_metadir(self._flow_name, run_id, step_name, task_id)
        selfname = os.path.join(subpath, '_self.json')
        self._makedirs(subpath)
        if os.path.isfile(selfname):
            return
        # In this case, the metadata information does not exist so we create it
        self._save_meta(
            subpath,
            {'_self': self._object_to_json(
                obj_type,
                run_id,
                step_name,
                task_id,
                tags + self.sticky_tags, sys_tags + self.sticky_sys_tags)})

    def _new_run(self, run_id, tags=[], sys_tags=[]):
        self._ensure_meta('flow', None, None, None)
        self._ensure_meta('run', run_id, None, None, tags, sys_tags)

    def _new_task(self, run_id, step_name, task_id, tags=[], sys_tags=[]):
        self._ensure_meta('step', run_id, step_name, None)
        self._ensure_meta('task', run_id, step_name, task_id, tags, sys_tags)
        self._register_code_package_metadata(run_id, step_name, task_id)

    @staticmethod
    def _make_path(
          flow_name=None, run_id=None, step_name=None, task_id=None, pathspec=None,
          create_on_absent=True):

        from metaflow.datastore.local import LocalDataStore
        if LocalDataStore.datastore_root is None:
            def print_clean(line, **kwargs):
                print(line)
            LocalDataStore.datastore_root = LocalDataStore.get_datastore_root_from_config(
                print_clean, create_on_absent=create_on_absent)
        if LocalDataStore.datastore_root is None:
            return None

        return LocalDataStore.make_path(flow_name, run_id, step_name, task_id, pathspec)

    @staticmethod
    def _create_and_get_metadir(
          flow_name=None, run_id=None, step_name=None, task_id=None):
        from metaflow.datastore.local import LocalDataStore
        root_path = LocalMetadataProvider._make_path(flow_name, run_id, step_name, task_id)
        subpath = os.path.join(root_path, LocalDataStore.METADATA_DIR)
        LocalMetadataProvider._makedirs(subpath)
        return subpath

    @staticmethod
    def _get_metadir(flow_name=None, run_id=None, step_name=None, task_id=None):
        from metaflow.datastore.local import LocalDataStore
        root_path = LocalMetadataProvider._make_path(
            flow_name, run_id, step_name, task_id, create_on_absent=False)
        if root_path is None:
            return None
        subpath = os.path.join(root_path, LocalDataStore.METADATA_DIR)
        if os.path.isdir(subpath):
            return subpath
        return None

    @staticmethod
    def _dump_json_to_file(
            filepath, data, allow_overwrite=False):
        if os.path.isfile(filepath) and not allow_overwrite:
            return
        with open(filepath + '.tmp', 'w') as f:
            json.dump(data, f)
        os.rename(filepath + '.tmp', filepath)

    @staticmethod
    def _read_json_file(filepath):
        with open(filepath, 'r') as f:
            return json.load(f)

    @staticmethod
    def _save_meta(root_dir, metadict):
        for name, datum in metadict.items():
            filename = os.path.join(root_dir, '%s.json' % name)
            LocalMetadataProvider._dump_json_to_file(filename, datum)
